from .PriorityMessage import PriorityMessage
from ..send import ISendService


class PrioritySendService(ISendService):
    def __init__(self, priority: int, send_service: ISendService) -> None:
        self.send_service = send_service
        self.priority = priority

    def send(self, message: bytes) -> None:
        self.send_service.send(PriorityMessage(self.priority, message).encode())



from __future__ import annotations
from logging import getLogger
from typing import List
from .FirstInLastOut import FirstInLastOut
from ..clock import IClock
from ..service import AbstractService, AbstractStartedService, AbstractServiceManager, StartedServiceWrapper


class MultiServiceManager(AbstractServiceManager):
    def __init__(self, clock: IClock, join_timeout_seconds: float = 0.1) -> None:
        super().__init__(clock)
        self._join_timeout_seconds = join_timeout_seconds
        self._logger = getLogger(__name__)
        self._not_started_services: FirstInLastOut[AbstractService] = FirstInLastOut()
        self._running_services: FirstInLastOut[AbstractStartedService] = FirstInLastOut()
        self._stopped_services: FirstInLastOut[AbstractStartedService] = FirstInLastOut()

    def add_service(self, service: AbstractService) -> None:
        self._not_started_services.append(service)

    def add_started_service(self, service: AbstractStartedService) -> None:
        self._running_services.append(service)

    def join_all(self) -> None:
        stubborn_services: List[AbstractStartedService] = []
        while True:
            service = self._stopped_services.pop()
            if service is None:
                break
            if service.join_service(self._join_timeout_seconds):
                stubborn_services.append(service)
        for service in stubborn_services:
            self._logger.info(f'Waiting for stubborn service "{service.get_service_name()}".')
            service.join_service()

    def start_all(self) -> None:
        while True:
            service = self._not_started_services.pop()
            if service is None:
                break
            service.start_service()
            StartedServiceWrapper(service, self)

    def stop_all(self) -> None:
        while True:
            service = self._running_services.pop()
            if service is None:
                break
            try:
                service.stop_service()
                self._stopped_services.append(service)
            except BaseException as exception:
                self._logger.exception(
                    f'Stopping service "{service.get_service_name()}" threw exception.', exc_info=exception)



from pigpio import pi
from .IGpio import IGpio


class PiGpio(IGpio):
    def __init__(self) -> None:
        self.pi = pi()

    def set_mode(self, gpio: int, mode: int) -> None:
        self.pi.set_mode(gpio, mode)

    def set_pwm_frequency(self, user_gpio: int, frequency: int) -> None:
        self.pi.set_PWM_frequency(user_gpio, frequency)

    def set_pwm_range(self, user_gpio: int, range_: int) -> None:
        self.pi.set_PWM_range(user_gpio, range_)

    def set_pwm_duty_cycle(self, user_gpio: int, duty_cycle: int) -> None:
        self.pi.set_PWM_dutycycle(user_gpio, duty_cycle)

    def stop(self) -> None:
        self.pi.stop()



from __future__ import annotations
from abc import ABC, abstractmethod
from .receive import IReceiveService
from .send import ISendService
from .util.ConnectionDetails import ConnectionDetails


class IUnreliableOsErrorListener(ABC):
    @abstractmethod
    def on_unreliable_os_error(self, os_error: OSError) -> None:
        """"""


class IUnreliableReceiveListener(ABC):
    @abstractmethod
    def on_unreliable_receive(self, message: bytes, details: ConnectionDetails) -> None:
        """"""


class IUnreliableService(ABC):
    @abstractmethod
    def add_unreliable_os_error_listener(self, listener: IUnreliableOsErrorListener) -> IUnreliableService:
        """"""

    @abstractmethod
    def add_unreliable_receive_listener(self, listener: IUnreliableReceiveListener) -> IUnreliableService:
        """"""


class IUnreliableReceiveSendService(IReceiveService, ISendService, IUnreliableService, ABC):
    """"""



from __future__ import annotations
from subprocess import PIPE, TimeoutExpired
from typing import Optional, Tuple
from .VideoRequestMessage import VideoRequestMessage
from .VideoResponseMessage import VideoResponseMessage
from .VideoSettings import VideoSettings
from ..argument import IArguments
from ..constants import caps_line_prefix
from ..message import message_types
from ..process import IProcess, ProcessFactory
from ..reliable import IReliableDisconnectListener, IReliableService
from ..receive import IReceiveListener
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..routed.RoutedSendService import RoutedSendService
from ..service import AbstractService, AbstractServiceManager
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Lazy import Lazy
from ..util.Placeholder import Placeholder


class GStreamerPipeline:
    def __init__(
            self,
            address: Tuple[str, int],
            caps: str,
            raspivid_process: IProcess,
            settings: VideoSettings
    ) -> None:
        self.address = address
        self.caps = caps
        self.raspivid_process = raspivid_process
        self.settings = settings


class GStreamerServerArguments(IArguments):
    def __init__(self) -> None:
        self.bin_gst_launch = Lazy(lambda store: 'gst-launch-1.0')
        self.bin_raspivid = Lazy(lambda store: 'raspivid')


class GStreamerServerService(
    AbstractService,
    IReceiveListener,
    IReliableDisconnectListener
):
    def __init__(
            self,
            arguments: GStreamerServerArguments,
            process_factory: ProcessFactory,
            routed_receive_service: RoutedReceiveService,
            reliable_service: IReliableService,
            routed_send_service: RoutedSendService,
            service_manager: AbstractServiceManager,
    ) -> None:
        super().__init__(service_manager)
        self.arguments = arguments
        self.caps: Placeholder[str] = Placeholder()
        self.gstreamer_pipeline: Placeholder[GStreamerPipeline] = Placeholder()
        self.process_factory = process_factory
        self.send_service = routed_send_service.create_send_service(message_types.video)
        routed_receive_service.create_receive_service(message_types.video).add_receive_listener(self)
        reliable_service.add_reliable_disconnect_listener(self)

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        gstreamer_pipeline = self.gstreamer_pipeline.get_optional()
        if gstreamer_pipeline is not None:
            try:
                gstreamer_pipeline.raspivid_process.wait(timeout_seconds)
                return False
            except TimeoutExpired:
                return True
        else:
            return False

    def open_video(self, address: Tuple[str, int], settings: VideoSettings) -> str:
        with self.gstreamer_pipeline as (gstreamer_pipeline, set_gstreamer_pipeline):
            if gstreamer_pipeline is not None:
                if gstreamer_pipeline.address == address \
                        and gstreamer_pipeline.settings.bit_rate == settings.bit_rate \
                        and gstreamer_pipeline.settings.fps == settings.fps \
                        and gstreamer_pipeline.settings.height == settings.height \
                        and gstreamer_pipeline.settings.width == settings.width:
                    return gstreamer_pipeline.caps
                else:
                    gstreamer_pipeline.raspivid_process.terminate()
            raspivid_process = self.process_factory((
                self.arguments.bin_raspivid.get(),
                '-n', '-t', '0',
                '-h', str(settings.height),
                '-w', str(settings.width),
                '-fps', str(settings.fps),
                '-b', str(settings.bit_rate),
                '-o', '-'
            ), None, PIPE)
            gst_launch_process = self.process_factory((
                self.arguments.bin_gst_launch.get(), '-v', 'fdsrc', '!', 'h264parse', '!', 'rtph264pay', '!', 'udpsink',
                f'host={address[0]}', f'port={address[1]}'
            ), raspivid_process.get_stdout(), PIPE)
            caps = None
            while caps is None:
                line = gst_launch_process.get_stdout().readline().decode()
                if line.startswith(caps_line_prefix):
                    caps = line[len(caps_line_prefix):-1]
            return set_gstreamer_pipeline(GStreamerPipeline(address, caps, raspivid_process, settings)).caps

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        request = VideoRequestMessage.decode(message)
        self.send_service.send(
            VideoResponseMessage(self.open_video(
                request.address,
                request.settings,
            ), request.address[1]).encode())

    def start_service(self) -> None:
        pass

    def stop_service(self) -> None:
        gstreamer_pipeline = self.gstreamer_pipeline.get_optional_and_clear()
        if gstreamer_pipeline is not None:
            gstreamer_pipeline.raspivid_process.terminate()

    def on_reliable_disconnect(self, details: ConnectionDetails) -> None:
        self.stop_service()



from __future__ import annotations
from abc import ABC, abstractmethod
from .ITimeoutReceiveListener import ITimeoutReceiveListener


class ITimeoutReceiveService(ABC):
    @abstractmethod
    def add_timeout_listener(self, timeout_listener: ITimeoutReceiveListener) -> ITimeoutReceiveService:
        """"""



from abc import ABC, abstractmethod


class IGpioServerService(ABC):
    @abstractmethod
    def update(self, motor_value: int, steering_value: int) -> None:
        """"""

    @abstractmethod
    def reset(self) -> None:
        """"""



from __future__ import annotations
from abc import ABC, abstractmethod
from .ILatencyListener import ILatencyListener


class ILatencyServerService(ABC):
    @abstractmethod
    def add_latency_listener(self, latency_listener: ILatencyListener) -> ILatencyServerService:
        """"""



from __future__ import annotations
from .RoutedMessage import RoutedMessage
from ..send import ISendService


class SendService(ISendService):
    def __init__(self, message_type: int, routed_send_service: RoutedSendService) -> None:
        self.message_type = message_type
        self.routed_send_service = routed_send_service

    def send(self, message: bytes) -> None:
        self.routed_send_service.send(self.message_type, message)


class RoutedSendService:
    def __init__(self, send_service: ISendService) -> None:
        self.send_service = send_service

    def create_send_service(self, message_type: int) -> ISendService:
        return SendService(message_type, self)

    def send(self, message_type: int, message: bytes) -> None:
        self.send_service.send(RoutedMessage(message_type, message).encode())



from __future__ import annotations
from logging import getLogger
from ..latency.ILatencyListener import ILatencyListener
from ..receive import IReceiveListener
from ..reliable import IReliableConnectListener, IReliableDisconnectListener, IReliableOsErrorListener
from ..reliable import IReliableReceiveListener
from ..routed.IRoutedReceiveListener import IRoutedReceiveListener
from ..unreliable import IUnreliableOsErrorListener, IUnreliableReceiveListener
from ..timeout.ITimeoutReceiveListener import ITimeoutReceiveListener
from ..util.ConnectionDetails import ConnectionDetails


class LogListener(
    ILatencyListener,
    IReceiveListener,
    IReliableConnectListener,
    IReliableDisconnectListener,
    IReliableOsErrorListener,
    IReliableReceiveListener,
    IRoutedReceiveListener,
    IUnreliableOsErrorListener,
    IUnreliableReceiveListener,
    ITimeoutReceiveListener,
):
    def __init__(self) -> None:
        self.logger = getLogger(__name__)

    def on_latency_available(self, latency: float) -> None:
        self.logger.info(f'on_latency_available({latency})')

    def on_latency_timeout(self) -> None:
        self.logger.info(f'on_latency_timeout()')

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.logger.info(f'on_received({message!r}, {details})')

    def on_routed_receive(self, message_type: int, message: bytes, details: ConnectionDetails) -> None:
        self.logger.info(f'on_routed_receive({message_type}, {message!r}, {details})')

    def on_timeout_receive(self) -> None:
        self.logger.info('on_receive_timeout')

    def on_reliable_connect(self, details: ConnectionDetails) -> None:
        self.logger.info(f'on_reliable_connected({details})')

    def on_reliable_disconnect(self, details: ConnectionDetails) -> None:
        self.logger.info(f'on_reliable_disconnected({details})')

    def on_reliable_os_error(self, os_error: OSError) -> None:
        self.logger.info(f'on_reliable_os_error({os_error})')

    def on_reliable_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.logger.info(f'on_reliable_received({message!r}, {details})')

    def on_unreliable_os_error(self, os_error: OSError) -> None:
        self.logger.info(f'on_unreliable_os_error({os_error})')

    def on_unreliable_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.logger.info(f'on_unreliable_received({message!r}, {details})')



from logging import getLogger
from socket import AF_INET, IPPROTO_UDP, SO_BROADCAST, SOCK_DGRAM, SOL_SOCKET
try:
    from socket import SO_REUSEPORT

    socket_can_reuse_port = True
except ImportError:
    SO_REUSEPORT = 0
    socket_can_reuse_port = False
from threading import Thread
from typing import Optional
from .DiscoveryCommonArguments import DiscoveryCommonArguments
from .DiscoveryRequestMessage import DiscoveryRequestMessage
from .DiscoveryResponseMessage import DiscoveryResponseMessage
from ..argument import IArguments
from ..constants import buffer_size
from ..log.util import log_method_call
from ..service import AbstractService, AbstractServiceManager
from ..socket_ import ISocket, ISocketFactory
from ..util.Placeholder import Placeholder


class DiscoveryServerArguments(IArguments):
    def __init__(self) -> None:
        self.common = DiscoveryCommonArguments()


class DiscoveryServerService(AbstractService):
    def __init__(
            self,
            arguments: DiscoveryServerArguments,
            listen_port: int,
            socket_factory: ISocketFactory,
            service_manager: AbstractServiceManager
    ) -> None:
        super().__init__(service_manager)
        logger = getLogger(__name__)
        self.arguments = arguments
        self.listen_port = listen_port
        self.should_run = True
        self.socket: Placeholder[ISocket] = Placeholder()
        self.socket_factory = socket_factory
        self.thread = Thread(target=log_method_call(logger, self.run))

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def run(self) -> None:
        with self.socket.set(self.socket_factory.socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) as socket:
            socket.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
            if socket_can_reuse_port:
                socket.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)
            socket.bind(('', self.arguments.common.port.get()))
            while self.should_run:
                message, address = socket.recvfrom(buffer_size)
                if address is not None and DiscoveryRequestMessage.is_valid(message):
                    response = DiscoveryResponseMessage(self.listen_port)
                    socket.sendto(response.encode(), address)

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        socket = self.socket.get_optional_and_clear()
        if socket is not None:
            socket.shutdown_guaranteed()



from __future__ import annotations
from typing import Tuple
from .VideoSettings import VideoSettings
from ..message import IMessage

separator = ','


class VideoRequestMessage(IMessage):
    def __init__(self, address: Tuple[str, int], settings: VideoSettings) -> None:
        self.address = address
        self.settings = settings

    def encode(self) -> bytes:
        return separator.join([
            self.address[0],
            str(self.address[1]),
            str(self.settings.bit_rate),
            str(self.settings.fps),
            str(self.settings.height),
            str(self.settings.width)
        ]).encode()

    @staticmethod
    def decode(message: bytes) -> VideoRequestMessage:
        ip, port, bit_rate, fps, height, width = message.decode().split(separator)
        return VideoRequestMessage((ip, int(port)), VideoSettings(int(bit_rate), int(fps), int(height), int(width)))



from __future__ import annotations
from logging import basicConfig, FileHandler, Formatter, Logger, StreamHandler
from typing import Callable, Optional, TypeVar

RV = TypeVar('RV')


def configure_log(log_file: Optional[str], log_level: int) -> None:
    formatter = Formatter(f'%(levelname)-5.5s:%(asctime)s:%(threadName)s:%(name)s:%(message)s')
    handlers = [StreamHandler()]
    if log_file is not None:
        handlers.append(FileHandler(log_file))
    for handler in handlers:
        handler.setFormatter(formatter)
    basicConfig(level=log_level, handlers=handlers)


def log_method_call(logger: Logger, method: Callable[[], RV]) -> Callable[[], RV]:
    def decorator() -> RV:
        try:
            logger.debug(f'entering {method.__name__}()')
            return method()
        finally:
            logger.debug(f'leaving {method.__name__}()')
    return decorator



from ..message import message_types
from ..receive import IReceiveListener
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..routed.RoutedSendService import RoutedSendService
from ..util.ConnectionDetails import ConnectionDetails


class LatencyClientService(IReceiveListener):
    def __init__(
            self,
            routed_receive_service: RoutedReceiveService,
            routed_send_service: RoutedSendService,
    ) -> None:
        self.send_service = routed_send_service.create_send_service(message_types.latency)
        routed_receive_service.create_receive_service(message_types.latency).add_receive_listener(self)

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.send_service.send(message)



from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

T = TypeVar('T')


class IQueue(Generic[T], ABC):
    @abstractmethod
    def empty(self) -> bool:
        """"""

    @abstractmethod
    def get(self, block: bool = True, timeout: Optional[float] = None) -> T:
        """"""

    @abstractmethod
    def join(self) -> None:
        """"""

    @abstractmethod
    def put(self, item: T) -> None:
        """"""



from types import TracebackType
from typing import Callable, Generic, Optional, Tuple, Type, TypeVar
from .Atomic import Atomic
from .FirstInLastOut import FirstInLastOut
from ..service import AbstractServiceManager
from ..promise.IPromise import IPromise
from ..promise.PromiseService import PromiseService

T = TypeVar('T')


class Placeholder(Generic[T]):
    def __init__(self, initial_value: Optional[T] = None) -> None:
        self._promises: FirstInLastOut[PromiseService[T]] = FirstInLastOut()
        self._value = Atomic(initial_value)

    def clear(self) -> None:
        with self._value as (_, set_optional):
            set_optional(None)

    def clear_and_set(self, next_optional: T) -> None:
        with self._value as (current_optional, set_optional):
            set_optional(None)
            _set(current_optional, next_optional, set_optional, self._promises)

    def get(self) -> T:
        with self._value as (optional, _):
            return _get(optional)

    def get_and_clear(self) -> T:
        with self._value as (optional, set_optional):
            return _get_and_clear(optional, set_optional)

    def get_eventually(self, service_manager: AbstractServiceManager) -> IPromise[T]:
        promise: PromiseService[T] = PromiseService(service_manager)
        with self._value as (optional, _):
            if optional is None:
                self._promises.append(promise)
            else:
                promise.resolve(optional)
        return promise

    def get_optional(self) -> Optional[T]:
        with self._value as (optional, _):
            return optional

    def get_optional_and_clear(self) -> Optional[T]:
        with self._value as (optional, set_optional):
            set_optional(None)
            return optional

    def is_present(self) -> bool:
        with self._value as (optional, _):
            return optional is not None

    def set(self, next_optional: T) -> T:
        with self._value as (current_optional, set_optional):
            _set(current_optional, next_optional, set_optional, self._promises)
        return next_optional

    def __enter__(self) -> Tuple[Optional[T], Callable[[Optional[T]], Optional[T]]]:
        return self._value.__enter__()

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        self._value.__exit__(exc_type, exc_val, exc_tb)


def _get(optional: Optional[T]) -> T:
    if optional is None:
        raise RuntimeError('Placeholder is empty')
    return optional


def _get_and_clear(
        optional: Optional[T],
        set_optional: Callable[[Optional[T]], Optional[T]],
) -> T:
    try:
        return _get(optional)
    finally:
        set_optional(None)


def _set(
        current_optional: Optional[T],
        next_optional: T,
        set_optional: Callable[[Optional[T]], Optional[T]],
        promises: FirstInLastOut[PromiseService[T]],
) -> None:
    if current_optional is not None:
        raise RuntimeError('Placeholder is not empty')
    set_optional(next_optional)
    while True:
        promise = promises.pop()
        if promise is None:
            break
        promise.resolve(next_optional)



from abc import ABC, abstractmethod


class IGpio(ABC):
    @abstractmethod
    def set_mode(self, gpio: int, mode: int) -> None:
        """"""

    @abstractmethod
    def set_pwm_frequency(self, user_gpio: int, frequency: int) -> None:
        """"""

    @abstractmethod
    def set_pwm_range(self, user_gpio: int, range_: int) -> None:
        """"""

    @abstractmethod
    def set_pwm_duty_cycle(self, user_gpio: int, duty_cycle: int) -> None:
        """"""

    @abstractmethod
    def stop(self) -> None:
        """"""



from abc import ABC, abstractmethod


class ILatencyListener(ABC):
    @abstractmethod
    def on_latency_available(self, latency: float) -> None:
        """"""

    @abstractmethod
    def on_latency_timeout(self) -> None:
        """"""



from __future__ import annotations
from threading import Event
from ..clock import IClock


class InterruptableSleep:
    """
    Inspired by https://stackoverflow.com/questions/24617131/is-there-a-graceful-or-pythonic-way-to-interrupt-a-time
    and https://stackoverflow.com/questions/16740104/python-lock-with-statement-and-timeout#16782391
    Also see https://docs.python.org/3/library/threading.html#threading.Event
    """
    def __init__(self, clock: IClock) -> None:
        self.clock = clock
        self.event = Event()

    def sleep(self, seconds: float) -> bool:
        """
        :return: True if slept through, False if sleep was interrupted.
        """
        self.event.clear()
        rv = not self.clock.wait_for_event(self.event, seconds)
        return rv

    def interrupt(self) -> None:
        self.event.set()



from __future__ import annotations
from .GStreamerVideoListener import GStreamerVideoListener
from .VideoRequestMessage import VideoRequestMessage
from .VideoResponseMessage import VideoResponseMessage
from .VideoSettings import VideoSettings
from ..argument import AnyArguments, IArguments, ValueArgument
from ..constants import default_video_port
from ..message import message_types
from ..routed.IRoutedReceiveListener import IRoutedReceiveListener
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..routed.RoutedSendService import RoutedSendService
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Lazy import Lazy
from ..util.Listeners import Listeners


class GStreamerClientArguments(IArguments):
    def __init__(self) -> None:
        self.flv_file_path = Lazy(lambda store: None)
        self.bit_rate = Lazy(lambda store: 2000000)
        self.fps = Lazy(lambda store: 30)
        self.height = Lazy(lambda store: 720)
        self.port = Lazy(lambda store: default_video_port)
        self.width = Lazy(lambda store: 1280)
        self.settings = Lazy(lambda store: VideoSettings(
            self.bit_rate(store),
            self.fps(store),
            self.height(store),
            self.width(store),
        ))

    def get_arguments(self) -> AnyArguments:
        return [
            ValueArgument(self.flv_file_path, '--flv-file-path', str, 'Store video stream additionally to this file.'),
            ValueArgument(self.bit_rate, '--video-bit-rate', int, 'Video quality parameter.'),
            ValueArgument(self.fps, '--video-fps', int, 'Video frames per second.'),
            ValueArgument(self.height, '--video-height', int, 'Video resolution height.'),
            ValueArgument(self.port, '--video-port', int, 'Video stream port.'),
            ValueArgument(self.width, '--video-width', int, 'Video resolution width.'),
        ]


class GStreamerClientService(IRoutedReceiveListener):
    def __init__(
            self,
            arguments: GStreamerClientArguments,
            routed_receive_service: RoutedReceiveService,
            routed_send_service: RoutedSendService,
    ) -> None:
        self.arguments = arguments
        self.routed_send_service = routed_send_service
        self.listeners: Listeners[GStreamerVideoListener] = Listeners()
        routed_receive_service.add_receive_listener(message_types.video, self)

    def add_gstreamer_video_listener(self, listener: GStreamerVideoListener) -> GStreamerClientService:
        self.listeners.add_listener(listener)
        return self

    def open_video(self, ip: str) -> None:
        self.routed_send_service.send(message_types.video, VideoRequestMessage(
            (ip, self.arguments.port.get()), self.arguments.settings.get()
        ).encode())

    def on_routed_receive(self, message_type: int, message: bytes, details: ConnectionDetails) -> None:
        response = VideoResponseMessage.decode(message)
        self.listeners.for_each(lambda listener: listener.on_video_available(response.caps, response.port))



from abc import ABC, abstractmethod
from threading import Event, Lock, Thread
from typing import Optional, TypeVar
from .queue_ import IQueue

T = TypeVar('T')


class IClock(ABC):
    @abstractmethod
    def acquire_lock(self, lock: Lock, timeout_seconds: Optional[float] = None) -> bool:
        """"""

    @abstractmethod
    def get_from_queue(self, queue: IQueue[T], timeout_seconds: Optional[float] = None) -> T:
        """"""

    @abstractmethod
    def get_seconds(self) -> float:
        """"""

    @abstractmethod
    def join_thread(self, thread: Thread, timeout_seconds: Optional[float] = None) -> None:
        """"""

    @abstractmethod
    def notify(self) -> None:
        """"""

    @abstractmethod
    def wait_for_event(self, event: Event, timeout_seconds: Optional[float] = None) -> bool:
        """"""



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Tuple
from .receive import IReceiveService
from .send import ISendService
from .util.ConnectionDetails import ConnectionDetails


class IReliableConnectListener(ABC):
    @abstractmethod
    def on_reliable_connect(self, details: ConnectionDetails) -> None:
        """"""


class IReliableDisconnectListener(ABC):
    @abstractmethod
    def on_reliable_disconnect(self, details: ConnectionDetails) -> None:
        """"""


class IReliableOsErrorListener(ABC):
    @abstractmethod
    def on_reliable_os_error(self, os_error: OSError) -> None:
        """"""


class IReliableReceiveListener(ABC):
    @abstractmethod
    def on_reliable_receive(self, message: bytes, details: ConnectionDetails) -> None:
        """"""


class IReliableService(ABC):
    @abstractmethod
    def add_reliable_connect_listener(self, listener: IReliableConnectListener) -> IReliableService:
        """"""

    @abstractmethod
    def add_reliable_disconnect_listener(self, listener: IReliableDisconnectListener) -> IReliableService:
        """"""

    @abstractmethod
    def add_reliable_os_error_listener(self, listener: IReliableOsErrorListener) -> IReliableService:
        """"""

    @abstractmethod
    def add_reliable_receive_listener(self, listener: IReliableReceiveListener) -> IReliableService:
        """"""

    @abstractmethod
    def get_own_address(self) -> Tuple[str, int]:
        """"""


class IReliableReceiveSendService(IReceiveService, IReliableService, ISendService, ABC):
    """"""



from socket import socket
from typing import Callable
from .ServerArguments import ServerArguments
from ..car.CarServerService import CarServerArguments, CarServerService
from ..clock import IClock
from ..discovery.DiscoveryServerService import DiscoveryServerArguments, DiscoveryServerService
from ..gpio.GpioServerService import GpioServerArguments, GpioServerService
from ..gpio.IGpio import IGpio
from ..gpio.IGpioServerService import IGpioServerService
from ..gpio.PiGpio import PiGpio
from ..gstreamer.GStreamerServerService import GStreamerServerArguments, GStreamerServerService
from ..latency.LatencyServerService import LatencyServerArguments, LatencyServerService
from ..log.LogArguments import LogArguments
from ..process import ProcessFactory
from ..reliable import IReliableReceiveSendService
from ..routed.RoutedSendService import RoutedSendService
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..service import AbstractServiceManager
from ..socket_ import ISocketFactory
from ..stop.StopServerService import StopServerService
from ..tcp.TcpServerService import TcpServerService
from ..udp.UdpService import UdpService
from ..unreliable import IUnreliableReceiveSendService
from ..util.Lazy import Lazy
from ..util.MultiServiceManager import MultiServiceManager
from ..util.Placeholder import Placeholder
from ..util.Process import Process
from ..util.RTC import rtc
from ..util.Socket import SocketFactory


class Server:
    def __init__(self) -> None:
        self.car_arguments = CarServerArguments()
        self.discovery_arguments = DiscoveryServerArguments()
        self.gpio_arguments = GpioServerArguments()
        self.gstreamer_arguments = GStreamerServerArguments()
        self.latency_arguments = LatencyServerArguments()
        self.log_arguments = LogArguments()
        self.server_arguments = ServerArguments()
        self.clock: Lazy[IClock] = Lazy(lambda store: rtc)
        self.process_factory: Lazy[ProcessFactory] = Lazy(lambda store: Process)
        self.socket_factory: Lazy[ISocketFactory] = Lazy(lambda store: SocketFactory(socket))
        self.service_manager: Lazy[AbstractServiceManager] = Lazy(lambda store: MultiServiceManager(self.clock(store)))
        self.discovery_service = Lazy(lambda store: DiscoveryServerService(
            self.discovery_arguments, self.server_arguments.listen_address(store)[1],
            self.socket_factory(store), self.service_manager(store)))
        self.gpio_factory: Lazy[Callable[[], IGpio]] = Lazy(lambda store: PiGpio)
        self.gpio_service: Lazy[IGpioServerService] = Lazy(lambda store: GpioServerService(
            self.gpio_arguments, self.gpio_factory(store), self.service_manager(store)))
        self.reliable_service: Lazy[IReliableReceiveSendService] = Lazy(lambda store: TcpServerService(
            self.server_arguments.listen_address(store), self.service_manager(store), self.socket_factory(store)))
        self.unreliable_service: Lazy[IUnreliableReceiveSendService] = Lazy(lambda store: UdpService(
            self.clock(store), True, Placeholder(self.server_arguments.listen_address(store)),
            self.service_manager(store), self.socket_factory(store),
            self.server_arguments.unreliable_timeout_seconds(store)))
        self.reliable_routed_receive_service = Lazy(lambda store: RoutedReceiveService(self.reliable_service(store)))
        self.reliable_routed_send_service = Lazy(lambda store: RoutedSendService(self.reliable_service(store)))
        self.unreliable_routed_receive_service = Lazy(lambda store: RoutedReceiveService(
            self.unreliable_service(store)))
        self.unreliable_routed_send_service = Lazy(lambda store: RoutedSendService(self.unreliable_service(store)))
        self.latency_service = Lazy(lambda store: LatencyServerService(
            self.latency_arguments, self.clock(store), self.unreliable_routed_receive_service(store),
            self.unreliable_routed_send_service(store), self.service_manager(store)))
        self.car_service = Lazy(lambda store: CarServerService(
            self.car_arguments, self.clock(store), self.gpio_service(store), self.latency_service(store),
            self.reliable_service(store), self.unreliable_routed_receive_service(store), self.service_manager(store)))
        self.gstreamer_service = Lazy(lambda store: GStreamerServerService(
            self.gstreamer_arguments, self.process_factory(store), self.reliable_routed_receive_service(store),
            self.reliable_service(store), self.reliable_routed_send_service(store), self.service_manager(store)))
        self.stop_service = Lazy(lambda store: StopServerService(
            self.reliable_routed_receive_service(store), self.service_manager(store)))



from logging import getLogger
from socket import AF_INET, IPPROTO_UDP, SO_BROADCAST, SOCK_DGRAM, SOL_SOCKET, timeout
try:
    from socket import SO_REUSEPORT

    socket_can_reuse_port = True
except ImportError:
    SO_REUSEPORT = 0
    socket_can_reuse_port = False
from threading import Thread
from typing import Optional, Tuple
from .DiscoveryCommonArguments import DiscoveryCommonArguments
from .DiscoveryRequestMessage import DiscoveryRequestMessage
from .DiscoveryResponseMessage import DiscoveryResponseMessage
from ..argument import AnyArguments, IArguments, ValueArgument
from ..constants import buffer_size
from ..log.util import log_method_call
from ..service import AbstractService, AbstractServiceManager
from ..socket_ import ISocket, ISocketFactory
from ..util.Lazy import Lazy
from ..util.Placeholder import Placeholder


class DiscoveryClientArguments(IArguments):
    def __init__(self) -> None:
        self.common = DiscoveryCommonArguments()
        self.ip = Lazy(lambda store: '255.255.255.255')
        self.address = Lazy(lambda store: (self.ip(store), self.common.port(store)))
        self.timeout_seconds = Lazy(lambda store: 2)

    def get_arguments(self) -> AnyArguments:
        return [
            ValueArgument(
                self.timeout_seconds, '--discovery-timeout-seconds', int,
                'Retry discovery after this amount of seconds.'),
            ValueArgument(self.ip, '--discovery-ip', str, 'Where to send discovery broadcast.'),
        ]


class DiscoveryClientService(AbstractService):
    def __init__(
            self,
            arguments: DiscoveryClientArguments,
            service_manager: AbstractServiceManager,
            socket_factory: ISocketFactory,
    ) -> None:
        super().__init__(service_manager)
        self.arguments = arguments
        self.logger = getLogger(__name__)
        self.server_address: Placeholder[Tuple[str, int]] = Placeholder()
        self.should_run = True
        self.socket: Placeholder[ISocket] = Placeholder()
        self.socket_factory = socket_factory
        self.thread = Thread(target=log_method_call(self.logger, self.run))

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def run(self) -> None:
        with self.socket.set(self.socket_factory.socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) as socket:
            socket.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
            if socket_can_reuse_port:
                socket.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)
            socket.settimeout(self.arguments.timeout_seconds.get())
            while self.should_run:
                try:
                    socket.sendto(DiscoveryRequestMessage.encode(), self.arguments.address.get())
                    message, address = socket.recvfrom(buffer_size)
                    if address is not None and DiscoveryResponseMessage.is_valid(message):
                        self.stop_service()
                        response = DiscoveryResponseMessage.decode(message)
                        self.server_address.set((address[0], response.port))
                except timeout:
                    self.logger.warning('timeout')
                except OSError as os_error:
                    if self.should_run:
                        raise os_error

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        socket = self.socket.get_optional_and_clear()
        if socket is not None:
            socket.shutdown_guaranteed()



from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

T = TypeVar('T')


class IPromise(Generic[T], ABC):
    @abstractmethod
    def get_blocking(self) -> Optional[T]:
        """"""



from __future__ import annotations
from abc import ABC, abstractmethod
from argparse import ArgumentParser, MetavarTypeHelpFormatter
from sys import version_info
from typing import Any, Dict, Generic, Iterable, Mapping, Optional, Sequence, Type, TypeVar
if version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict
from .util.checking import check_type
from .util.Lazy import Lazy

K = TypeVar('K')
T = TypeVar('T')
ParsedType = TypeVar('ParsedType')


def get_key_of_value(haystack: Mapping[K, T], needle: T) -> K:
    for key, value in haystack.items():
        if needle == value:
            return key
    raise RuntimeError(f'Value "{needle}" not found in mapping.')


class AddArgumentKwargs(TypedDict, total=False):
    choices: Iterable[Any]
    dest: str
    help: str
    type: Type[Any]


class IArguments(ABC):
    def get_arguments(self) -> AnyArguments:
        return []


class AbstractArgument(Generic[T, ParsedType], IArguments, ABC):
    def __init__(
            self, lazy: Lazy[T], name: str, type_: Type[ParsedType], help_: str = '', default: Optional[str] = ''
    ) -> None:
        self.help = help_ if default is None else f'{help_} (default: {default or lazy.get_default()})'
        self.lazy = lazy
        self.name = name
        self.type = type_

    def get_kwargs_for_add_argument(self) -> AddArgumentKwargs:
        return dict()

    def add_argument(self, parser: ArgumentParser) -> None:
        kwargs = self.get_kwargs_for_add_argument()
        if self.name[0] in parser.prefix_chars:
            kwargs['dest'] = self.name
        kwargs['help'] = self.help
        kwargs['type'] = self.type
        parser.add_argument(self.name, **kwargs)

    @abstractmethod
    def parse_value(self, value: ParsedType) -> T:
        """"""

    def get_arguments(self) -> AnyArguments:
        return [self]


AnyArgument = AbstractArgument[Any, Any]
AnyArguments = Iterable[AnyArgument]


class ChoiceArgument(Generic[T], AbstractArgument[T, str]):
    def __init__(self, lazy: Lazy[T], name: str, choices: Mapping[str, T], help_: str = '', default: str = '') -> None:
        super().__init__(lazy, name, str, help_, default or get_key_of_value(choices, lazy.get_default()))
        self.choices = choices

    def get_kwargs_for_add_argument(self) -> AddArgumentKwargs:
        return dict(choices=self.choices)

    def parse_value(self, value: str) -> T:
        return self.choices[check_type(value, self.type)]


class ValueArgument(Generic[T], AbstractArgument[T, T]):
    def parse_value(self, value: T) -> T:
        return check_type(value, self.type)


class ArgumentRegistry:
    def __init__(self) -> None:
        self.parser = ArgumentParser(formatter_class=MetavarTypeHelpFormatter)
        self.arguments: Dict[str, AbstractArgument[Any, Any]] = {}

    def add_arguments(self, arguments_list: Iterable[IArguments]) -> None:
        for arguments in arguments_list:
            for argument in arguments.get_arguments():
                self.add_argument(argument)

    def add_argument(self, argument: AbstractArgument[T, ParsedType]) -> None:
        name = argument.name
        if name in self.arguments:
            raise RuntimeError(f'Argument {name} already used.')
        self.arguments[name] = argument
        argument.add_argument(self.parser)

    def parse_arguments(self, argument_vector: Optional[Sequence[str]] = None) -> None:
        parsed = vars(self.parser.parse_args(argument_vector))
        parsed_keys = set(parsed.keys())
        for name, value in parsed.items():
            if name not in self.arguments:
                raise RuntimeError(f'Unknown argument {name}.')
            if value is not None:
                argument = self.arguments[name]
                argument.lazy.set(argument.parse_value(value))
            parsed_keys.remove(name)
        if len(parsed_keys) != 0:
            unknown_arguments_string = ', '.join((f'{key}: {parsed[key]}' for key in parsed_keys))
            raise RuntimeError(f'Unknown arguments: {unknown_arguments_string}')
        for argument in self.arguments.values():
            argument.lazy.get()


def process_arguments(arguments_list: Iterable[IArguments], argument_vector: Optional[Sequence[str]] = None) -> None:
    argument_registry = ArgumentRegistry()
    argument_registry.add_arguments(arguments_list)
    argument_registry.parse_arguments(argument_vector)



from __future__ import annotations
from struct import pack, unpack
from typing import cast, Tuple
from ..message import encode_payload, IMessage, Payload
from ..util.struct import get_required_size

format_string = '<B'
format_length = get_required_size(format_string)
format_type = Tuple[int]


class RoutedMessage(IMessage):
    def __init__(self, message_type: int, payload: Payload) -> None:
        self.message_type = message_type
        self.payload = encode_payload(payload)

    def encode(self) -> bytes:
        return pack(format_string, self.message_type) + self.payload

    @staticmethod
    def decode(message: bytes) -> RoutedMessage:
        message_type, = cast(format_type, unpack(format_string, message[:format_length]))
        return RoutedMessage(message_type, message[format_length:])



from __future__ import annotations
from struct import pack, unpack
from typing import cast, Tuple
from ..message import IMessage
from ..util.struct import get_required_size

format_string = '<H'
format_length = get_required_size(format_string)
format_type = Tuple[int]


class VideoResponseMessage(IMessage):
    def __init__(self, caps: str, port: int) -> None:
        self.caps = caps
        self.port = port

    def encode(self) -> bytes:
        return pack(format_string, self.port) + self.caps.encode()

    @staticmethod
    def decode(message: bytes) -> VideoResponseMessage:
        port, = cast(format_type, unpack(format_string, message[:format_length]))
        caps = message[format_length:].decode()
        return VideoResponseMessage(caps, port)



from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from ..argument import AnyArguments, ChoiceArgument, IArguments, ValueArgument
from ..util.Lazy import Lazy


class LogArguments(IArguments):
    def __init__(self) -> None:
        self.log_file = Lazy(lambda store: None)
        self.log_level = Lazy(lambda store: WARNING)

    def get_arguments(self) -> AnyArguments:
        return [
            ValueArgument(
                self.log_file, '--log-file', str, 'Gets created if does not exists. Gets appended if exists.'),
            ChoiceArgument(self.log_level, '--log-level', {
                'debug': DEBUG,
                'info': INFO,
                'warn': WARNING,
                'error': ERROR,
                'critical': CRITICAL,
            }, 'Verbosity of the log output.')
        ]



from __future__ import annotations
from threading import Thread
from typing import Optional
from .PriorityMessage import PriorityMessage
from ..clock import IClock
from ..service import AbstractService, AbstractServiceManager
from ..receive import IReceiveListener, IReceiveService
from ..util.Atomic import Atomic
from ..util.ConnectionDetails import ConnectionDetails
from ..util.InterruptableSleep import InterruptableSleep
from ..util.Listeners import Listeners


class PriorityReceiveService(AbstractService, IReceiveListener, IReceiveService):
    def __init__(
            self,
            clock: IClock,
            receive_service: IReceiveService,
            service_manager: AbstractServiceManager,
            timeout_seconds: float,
    ) -> None:
        super().__init__(service_manager)
        self.current_priority = Atomic(0)
        self.interruptable_sleep = InterruptableSleep(clock)
        self.listeners: Listeners[IReceiveListener] = Listeners()
        self.should_run = True
        self.thread = Thread(target=self.run)
        self.timeout_seconds = timeout_seconds
        receive_service.add_receive_listener(self)

    def add_receive_listener(self, listener: IReceiveListener) -> PriorityReceiveService:
        self.listeners.add_listener(listener)
        return self

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.interruptable_sleep.interrupt()
        priority_message = PriorityMessage.decode(message)
        with self.current_priority as (current_priority, set_current_priority):
            if priority_message.priority > current_priority:
                set_current_priority(priority_message.priority)
            elif priority_message.priority < current_priority:
                return
        self.listeners.for_each(lambda listener: listener.on_receive(priority_message.payload, details))

    def run(self) -> None:
        while self.should_run:
            if self.interruptable_sleep.sleep(self.timeout_seconds) and self.should_run:
                with self.current_priority as (_, set_current_priority):
                    set_current_priority(0)

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        self.interruptable_sleep.interrupt()



from __future__ import annotations
from socket import socket
from types import TracebackType
from typing import Callable, Optional, Tuple, Type, Union
from ..socket_ import ISocket, ISocketFactory
from ..util.checking import check_type_tuple2


class Socket(ISocket):
    def __init__(self, socket_: socket) -> None:
        self.socket = socket_

    def __enter__(self) -> Socket:
        return Socket(self.socket.__enter__())

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        self.socket.__exit__(exc_type, exc_val, exc_tb)

    def accept(self) -> Tuple[Socket, Tuple[str, int]]:
        socket_, address = self.socket.accept()
        return Socket(socket_), address

    def bind(self, address: Tuple[str, int]) -> None:
        self.socket.bind(address)

    def close(self) -> None:
        self.socket.close()

    def connect(self, address: Tuple[str, int]) -> None:
        self.socket.connect(address)

    def getsockname(self) -> Tuple[str, int]:
        return check_type_tuple2(self.socket.getsockname(), str, int)

    def listen(self, backlog: int) -> None:
        self.socket.listen(backlog)

    def recv(self, bufsize: int) -> bytes:
        return self.socket.recv(bufsize)

    def recvfrom(self, bufsize: int) -> Tuple[bytes, Tuple[str, int]]:
        return self.socket.recvfrom(bufsize)

    def sendall(self, data: bytes) -> None:
        self.socket.sendall(data)

    def sendto(self, data: bytes, address: Tuple[str, int]) -> int:
        return self.socket.sendto(data, address)

    def setsockopt(self, level: int, optname: int, value: Union[int, bytes]) -> None:
        self.socket.setsockopt(level, optname, value)

    def settimeout(self, value: float) -> None:
        self.socket.settimeout(value)

    def shutdown(self, how: int) -> None:
        self.socket.shutdown(how)


class SocketFactory(ISocketFactory):
    def __init__(self, socket_factory: Callable[[int, int, int], socket]) -> None:
        self.socket_factory = socket_factory

    def socket(self, family: int, type_: int, proto: int = -1) -> Socket:
        return Socket(self.socket_factory(family, type_, proto))



from __future__ import annotations
from logging import getLogger
from .Placeholder import Placeholder
from ..clock import IClock
from ..service import AbstractService, AbstractStartedService, AbstractServiceManager, StartedServiceWrapper


class SingleServiceManager(AbstractServiceManager):
    def __init__(self, clock: IClock) -> None:
        super().__init__(clock)
        self.logger = getLogger(__name__)
        self._not_started_service: Placeholder[AbstractService] = Placeholder()
        self._running_service: Placeholder[AbstractStartedService] = Placeholder()
        self._stopped_service: Placeholder[AbstractStartedService] = Placeholder()

    def add_service(self, service: AbstractService) -> None:
        self._not_started_service.set(service)

    def add_started_service(self, service: AbstractStartedService) -> None:
        self._running_service.set(service)

    def is_service_running(self) -> bool:
        return self._running_service.is_present()

    def join_all(self) -> None:
        service = self._stopped_service.get_optional_and_clear()
        if service is not None:
            service.join_service()

    def start_all(self) -> None:
        service = self._not_started_service.get_and_clear()
        service.start_service()
        StartedServiceWrapper(service, self)

    def stop_all(self) -> None:
        service = self._running_service.get_optional_and_clear()
        if service is not None:
            try:
                service.stop_service()
                self._stopped_service.set(service)
            except BaseException as exception:
                self.logger.exception(f'Stopping service "{type(service)}" threw exception.', exc_info=exception)



from __future__ import annotations
from typing import Callable, Optional
from .IGpio import IGpio
from .IGpioServerService import IGpioServerService
from ..argument import IArguments
from ..service import AbstractService, AbstractServiceManager
from ..util.Placeholder import Placeholder
from ..util.util import min_max
from ..util.Lazy import Lazy


class GpioServerArguments(IArguments):
    def __init__(self) -> None:
        self.pwm_base = Lazy(lambda store: 300)
        self.pwm_frequency = Lazy(lambda store: 50)
        self.pwm_range = Lazy(lambda store: 4000)
        self.pwm_motor_base = Lazy(lambda store: self.pwm_base(store))
        self.pwm_motor_frequency = Lazy(lambda store: self.pwm_frequency(store))
        self.pwm_motor_gpio = Lazy(lambda store: 12)
        self.pwm_motor_invert = Lazy(lambda store: False)
        self.pwm_motor_maximum = Lazy(lambda store: 100)
        self.pwm_motor_minimum = Lazy(lambda store: -self.pwm_motor_maximum(store))
        self.pwm_motor_offset = Lazy(lambda store: 0)
        self.pwm_motor_range = Lazy(lambda store: self.pwm_range(store))
        self.pwm_steering_base = Lazy(lambda store: self.pwm_base(store))
        self.pwm_steering_frequency = Lazy(lambda store: self.pwm_frequency(store))
        self.pwm_steering_gpio = Lazy(lambda store: 13)
        self.pwm_steering_invert = Lazy(lambda store: False)
        self.pwm_steering_maximum = Lazy(lambda store: 100)
        self.pwm_steering_minimum = Lazy(lambda store: -self.pwm_steering_maximum(store))
        self.pwm_steering_offset = Lazy(lambda store: 0)
        self.pwm_steering_range = Lazy(lambda store: self.pwm_range(store))


class GpioServerService(IGpioServerService, AbstractService):
    def __init__(
            self, arguments: GpioServerArguments, gpio_factory: Callable[[], IGpio],
            service_manager: AbstractServiceManager,
    ) -> None:
        super().__init__(service_manager)
        self.arguments = arguments
        self.gpio_factory = gpio_factory
        self.gpio: Placeholder[IGpio] = Placeholder()

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        return False

    def start_service(self) -> None:
        gpio = self.gpio_factory()
        gpio.set_mode(self.arguments.pwm_motor_gpio.get(), 1)
        gpio.set_pwm_frequency(self.arguments.pwm_motor_gpio.get(), self.arguments.pwm_motor_frequency.get())
        gpio.set_pwm_range(self.arguments.pwm_motor_gpio.get(), self.arguments.pwm_motor_range.get())
        _set_motor(self.arguments, gpio, 0)
        gpio.set_mode(self.arguments.pwm_steering_gpio.get(), 1)
        gpio.set_pwm_frequency(self.arguments.pwm_steering_gpio.get(), self.arguments.pwm_steering_frequency.get())
        gpio.set_pwm_range(self.arguments.pwm_steering_gpio.get(), self.arguments.pwm_steering_range.get())
        _set_steering(self.arguments, gpio, 0)
        self.gpio.set(gpio)

    def update(self, motor_value: int, steering_value: int) -> None:
        with self.gpio as (gpio, _):
            if gpio is not None:
                _set_motor(self.arguments, gpio, motor_value)
                _set_steering(self.arguments, gpio, steering_value)

    def reset(self) -> None:
        self.update(0, 0)

    def stop_service(self) -> None:
        self.reset()
        self.gpio.get_and_clear().stop()


def _set_motor(arguments: GpioServerArguments, gpio: IGpio, value: int) -> None:
    factor = -1 if arguments.pwm_motor_invert.get() else 1
    pwm_value = arguments.pwm_motor_base.get() + arguments.pwm_motor_offset.get() + min_max(
        value * factor, arguments.pwm_motor_minimum.get(), arguments.pwm_motor_maximum.get())
    gpio.set_pwm_duty_cycle(arguments.pwm_motor_gpio.get(), pwm_value)


def _set_steering(arguments: GpioServerArguments, gpio: IGpio, value: int) -> None:
    factor = -1 if arguments.pwm_steering_invert.get() else 1
    pwm_value = arguments.pwm_steering_base.get() + arguments.pwm_steering_offset.get() + min_max(
        value * factor, arguments.pwm_steering_minimum.get(), arguments.pwm_steering_maximum.get())
    gpio.set_pwm_duty_cycle(arguments.pwm_steering_gpio.get(), pwm_value)



from __future__ import annotations
from struct import pack, unpack
from typing import cast, Tuple
from ..message import IMessage
from ..constants import discovery_magic_response_bytes

format_string = '<H'
format_type = Tuple[int]


class DiscoveryResponseMessage(IMessage):
    def __init__(self, port: int) -> None:
        self.port = port

    def encode(self) -> bytes:
        return discovery_magic_response_bytes + pack(format_string, self.port)

    @staticmethod
    def decode(message: bytes) -> DiscoveryResponseMessage:
        magic_bytes_length = len(discovery_magic_response_bytes)
        port, = cast(format_type, unpack(format_string, message[magic_bytes_length:]))
        return DiscoveryResponseMessage(port)

    @staticmethod
    def is_valid(message: bytes) -> bool:
        return message.startswith(discovery_magic_response_bytes)



from abc import ABC, abstractmethod
from typing import Callable, IO, Optional, Sequence, Union


class IProcess(ABC):
    @abstractmethod
    def get_stdout(self) -> IO[bytes]:
        """"""

    @abstractmethod
    def terminate(self) -> None:
        """"""

    @abstractmethod
    def wait(self, timeout: Optional[float]) -> None:
        """"""


FILE = Union[int, IO[bytes]]
ProcessFactory = Callable[[Sequence[str], Optional[FILE], Optional[FILE]], IProcess]



from __future__ import annotations
from threading import Thread
from typing import Callable, Optional
from ..clock import IClock
from ..service import AbstractService, AbstractServiceManager
from ..send import ISendService
from ..util.Atomic import Atomic
from ..util.AtomicInteger import AtomicInteger
from ..util.InterruptableSleep import InterruptableSleep


class TimeoutSendService(AbstractService):
    def __init__(
            self,
            clock: IClock,
            timeout_seconds: float,
            send_service: ISendService,
            service_manager: AbstractServiceManager,
            message_callback: Callable[[], Optional[bytes]] = lambda: None,
    ) -> None:
        super().__init__(service_manager)
        self.interruptable_sleep = InterruptableSleep(clock)
        self.timeout_seconds = timeout_seconds
        self.message_callback = Atomic(message_callback)
        self.send_count = AtomicInteger(0)
        self.send_service = send_service
        self.should_run = True
        self.thread = Thread(target=self.run)

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def run(self) -> None:
        should_send = True
        while self.should_run:
            if should_send:
                with self.message_callback as (message_callback, _):
                    self._send(message_callback())
            should_send = self.interruptable_sleep.sleep(self.timeout_seconds)

    def set_and_send_immediately(self, message_callback: Callable[[], Optional[bytes]]) -> Optional[bytes]:
        with self.message_callback as (_, set_message_callback):
            message = message_callback()
            self._send(message)
            set_message_callback(message_callback)
        self.interruptable_sleep.interrupt()
        return message

    def _send(self, message: Optional[bytes]) -> None:
        if message is not None:
            self.send_count.increment()
            self.send_service.send(message)

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        self.interruptable_sleep.interrupt()



from __future__ import annotations
from abc import ABC, abstractmethod
from types import TracebackType
from typing import Optional, Type
from .clock import IClock


class AbstractService(ABC):
    def __init__(self, service_manager: AbstractServiceManager) -> None:
        service_manager.add_service(self)

    @abstractmethod
    def get_service_name(self) -> str:
        """"""

    @abstractmethod
    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        """
        :return: True if services is still running, False otherwise.
        """

    @abstractmethod
    def start_service(self) -> None:
        """"""

    @abstractmethod
    def stop_service(self) -> None:
        """"""


class AbstractStartedService(ABC):
    def __init__(self, service_manager: AbstractServiceManager) -> None:
        service_manager.add_started_service(self)

    @abstractmethod
    def get_service_name(self) -> str:
        """"""

    @abstractmethod
    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        """"""

    @abstractmethod
    def stop_service(self) -> None:
        """"""


class StartedServiceWrapper(AbstractStartedService):
    def __init__(self, service: AbstractService, service_manager: AbstractServiceManager) -> None:
        super().__init__(service_manager)
        self.service = service

    def get_service_name(self) -> str:
        return self.service.get_service_name()

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        return self.service.join_service(timeout_seconds)

    def stop_service(self) -> None:
        self.service.stop_service()


class AbstractServiceManager(ABC):
    def __init__(self, clock: IClock) -> None:
        self.clock = clock

    @abstractmethod
    def add_service(self, service: AbstractService) -> None:
        """"""

    @abstractmethod
    def add_started_service(self, service: AbstractStartedService) -> None:
        """"""

    @abstractmethod
    def join_all(self) -> None:
        """"""

    @abstractmethod
    def start_all(self) -> None:
        """"""

    @abstractmethod
    def stop_all(self) -> None:
        """"""

    def __enter__(self) -> None:
        self.start_all()

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        self.stop_all()
        self.clock.notify()
        self.join_all()



from __future__ import annotations
from logging import getLogger
from socket import AF_INET, SOCK_STREAM
from threading import Thread
from types import TracebackType
from typing import Optional, Tuple, Type
from ..constants import buffer_size
from ..log.util import log_method_call
from ..receive import IReceiveListener
from ..reliable import IReliableReceiveSendService, IReliableOsErrorListener, IReliableReceiveListener
from ..reliable import IReliableConnectListener, IReliableDisconnectListener
from ..service import AbstractService, AbstractServiceManager
from ..socket_ import ISocket, ISocketFactory
from ..util import checking
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Listeners import Listeners
from ..util.Placeholder import Placeholder


class AcceptWrapper:
    def __init__(self, connection: Tuple[ISocket, Tuple[str, int]]) -> None:
        self.socket = connection[0]
        self.address = connection[1]

    def __enter__(self) -> Tuple[ISocket, Tuple[str, int]]:
        return self.socket.__enter__(), self.address

    def set_into_placeholder(self, placeholder: Placeholder[ISocket]) -> AcceptWrapper:
        placeholder.set(self.socket)
        return self

    def __exit__(
            self, exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        self.socket.__exit__(exc_type, exc_val, exc_tb)


class TcpServerService(AbstractService, IReliableReceiveSendService):
    def __init__(
            self,
            listen_address: Tuple[str, int],
            service_manager: AbstractServiceManager,
            socket_factory: ISocketFactory,
    ) -> None:
        super().__init__(service_manager)
        self.connected_listeners: Listeners[IReliableConnectListener] = Listeners()
        self.disconnected_listeners: Listeners[IReliableDisconnectListener] = Listeners()
        self.listen_address = listen_address
        self.logger = getLogger(__name__)
        self.os_error_listeners: Listeners[IReliableOsErrorListener] = Listeners()
        self.received_listeners: Listeners[IReceiveListener] = Listeners()
        self.reliable_received_listeners: Listeners[IReliableReceiveListener] = Listeners()
        self.should_run = True
        self.socket_factory = socket_factory
        self.server_socket: Placeholder[ISocket] = Placeholder()
        self.socket: Placeholder[ISocket] = Placeholder()
        self.thread = Thread(target=log_method_call(self.logger, self.run))

    def add_reliable_connect_listener(self, listener: IReliableConnectListener) -> TcpServerService:
        self.connected_listeners.add_listener(listener)
        return self

    def add_reliable_disconnect_listener(self, listener: IReliableDisconnectListener) -> TcpServerService:
        self.disconnected_listeners.add_listener(listener)
        return self

    def add_receive_listener(self, listener: IReceiveListener) -> TcpServerService:
        self.received_listeners.add_listener(listener)
        return self

    def add_reliable_os_error_listener(self, listener: IReliableOsErrorListener) -> TcpServerService:
        self.os_error_listeners.add_listener(listener)
        return self

    def add_reliable_receive_listener(
            self,
            listener: IReliableReceiveListener
    ) -> TcpServerService:
        self.reliable_received_listeners.add_listener(listener)
        return self

    def get_service_name(self) -> str:
        return __name__

    def get_own_address(self) -> Tuple[str, int]:
        return self.listen_address

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def run_connected_client(self, socket: ISocket, connection_details: ConnectionDetails) -> None:
        self.connected_listeners.for_each(
            lambda listener: listener.on_reliable_connect(connection_details))
        try:
            while self.should_run:
                message = socket.recv(buffer_size)
                if len(message) == 0:  # ... means that peer has gracefully shutdown connection
                    break
                self.received_listeners.for_each(
                    lambda listener: listener.on_receive(message, connection_details))
                self.reliable_received_listeners.for_each(
                    lambda listener: listener.on_reliable_receive(message, connection_details))
        except OSError as os_error:
            self.os_error_listeners.for_each(
                lambda listener: listener.on_reliable_os_error(os_error))
        finally:
            socket.shutdown_guaranteed()
            self.disconnected_listeners.for_each(
                lambda listener: listener.on_reliable_disconnect(connection_details))

    def run(self) -> None:
        with self.server_socket.set(self.socket_factory.socket(AF_INET, SOCK_STREAM)) as server_socket:
            server_socket.bind(self.listen_address)
            server_socket.listen(1)
            while self.should_run:
                try:
                    with AcceptWrapper(server_socket.accept()).set_into_placeholder(self.socket) as (
                            socket, client_address):
                        self.run_connected_client(socket, ConnectionDetails(socket.getsockname(), client_address))
                except OSError as os_error:
                    if self.should_run:
                        raise os_error
                if self.should_run:
                    self.socket.clear()

    def send(self, message: bytes) -> None:
        checking.check_message_length(len(message))
        self.socket.get().sendall(message)

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        socket = self.socket.get_optional()
        if socket is not None:
            socket.shutdown_guaranteed()
        server_socket = self.server_socket.get_optional_and_clear()
        if server_socket is not None:
            server_socket.shutdown_guaranteed()



from logging import getLogger
from threading import Event
from typing import Optional
from ..service import AbstractStartedService, AbstractServiceManager
from ..routed.IRoutedReceiveListener import IRoutedReceiveListener
from ..message import message_types
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..util.ConnectionDetails import ConnectionDetails


class StopServerService(AbstractStartedService, IRoutedReceiveListener):
    def __init__(
            self,
            routed_receive_service: RoutedReceiveService,
            service_manager: AbstractServiceManager
    ) -> None:
        super().__init__(service_manager)
        self.event = Event()
        self.logger = getLogger(__name__)
        routed_receive_service.add_receive_listener(message_types.stop, self)

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        return not self.event.is_set()

    def on_routed_receive(self, message_type: int, message: bytes, details: ConnectionDetails) -> None:
        self.event.set()

    def wait(self) -> None:
        try:
            self.event.wait()
        except KeyboardInterrupt:
            self.logger.info('terminating . . .')

    def stop_service(self) -> None:
        self.event.set()



from __future__ import annotations
from .ExpireMessage import ExpireMessage, highest_message_number, message_number_range, valid_message_number_range
from .IExpireReceiveListener import IExpireReceiveListener
from ..receive import IReceiveListener, IReceiveService
from ..util.Atomic import Atomic
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Listeners import Listeners


class ExpireReceiveService(IReceiveListener, IReceiveService):
    def __init__(self, receive_service: IReceiveService) -> None:
        self.last_message_number = Atomic(0)
        self.expire_listeners: Listeners[IExpireReceiveListener] = Listeners()
        self.receive_listeners: Listeners[IReceiveListener] = Listeners()
        receive_service.add_receive_listener(self)

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        expire_message = ExpireMessage.decode(message)
        with self.last_message_number as (last_message_number, set_last_message_number):
            end_of_valid_range = (last_message_number + valid_message_number_range) % message_number_range
            if expire_message.message_number == highest_message_number:
                is_valid = True
            elif end_of_valid_range < last_message_number:
                is_valid = not end_of_valid_range <= expire_message.message_number <= last_message_number
            else:
                is_valid = last_message_number < expire_message.message_number < end_of_valid_range
            if is_valid:
                set_last_message_number(expire_message.message_number)
        if is_valid:
            self.receive_listeners.for_each(lambda listener: listener.on_receive(expire_message.payload, details))
        else:
            self.expire_listeners.for_each(lambda listener: listener.on_expire_receive(expire_message.payload, details))

    def add_receive_listener(self, listener: IReceiveListener) -> ExpireReceiveService:
        self.receive_listeners.add_listener(listener)
        return self

    def add_expire_listener(self, listener: IExpireReceiveListener) -> ExpireReceiveService:
        self.expire_listeners.add_listener(listener)
        return self



from __future__ import annotations
from ..message import IMessage

separator = ','


class NetworkStatisticsMessage(IMessage):
    def __init__(self, expire_receive_count: int, receive_count: int, timeout_receive_count: int) -> None:
        self.expire_receive_count = expire_receive_count
        self.receive_count = receive_count
        self.timeout_receive_count = timeout_receive_count

    def encode(self) -> bytes:
        return separator.join([
            str(self.expire_receive_count),
            str(self.receive_count),
            str(self.timeout_receive_count),
        ]).encode()

    @staticmethod
    def decode(message: bytes) -> NetworkStatisticsMessage:
        expire_receive_count, receive_count, timeout_receive_count = message.decode().split(separator)
        return NetworkStatisticsMessage(int(expire_receive_count), int(receive_count), int(timeout_receive_count))



"""
see https://docs.python.org/3/library/struct.html#format-characters
"""
size_map = {
    'c': 1,
    'b': 1,
    'B': 1,
    '?': 1,
    'h': 2,
    'H': 2,
    'i': 4,
    'I': 4,
    'l': 4,
    'L': 4,
    'q': 8,
    'Q': 8,
    'e': 2,
    'f': 4,
    'd': 8,
}


def get_required_size(format_string: str) -> int:
    format_list = list(format_string)
    size = 0
    if format_list[0] == '<':
        format_list.pop(0)
    while len(format_list) != 0:
        size += size_map[format_list.pop(0)]
    return size



from __future__ import annotations
from struct import pack, unpack
from typing import cast, Tuple
from ..message import IMessage

format_string = '<B'
format_type = Tuple[int]
max_number = 2**8


class LatencyMessage(IMessage):
    def __init__(self, number: int) -> None:
        self.number = number

    def encode(self) -> bytes:
        return pack(format_string, self.number)

    def increment(self) -> None:
        self.number = (self.number + 1) % max_number

    @staticmethod
    def decode(message: bytes) -> LatencyMessage:
        number, = cast(format_type, unpack(format_string, message))
        return LatencyMessage(number)



from __future__ import annotations
from queue import Empty
from threading import Thread
from typing import Optional
from .ILatencyListener import ILatencyListener
from .ILatencyServerService import ILatencyServerService
from .LatencyMessage import LatencyMessage
from ..argument import IArguments
from ..clock import IClock
from ..message import message_types
from ..queue_ import IQueue
from ..receive import IReceiveListener
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..routed.RoutedSendService import RoutedSendService
from ..service import AbstractService, AbstractServiceManager
from ..util.ConnectionDetails import ConnectionDetails
from ..util.InterruptableSleep import InterruptableSleep
from ..util.Lazy import Lazy
from ..util.IdealQueue import IdealQueue
from ..util.Listeners import Listeners


class LatencyServerArguments(IArguments):
    def __init__(self) -> None:
        self.latency_check_interval_seconds = Lazy(lambda store: 1.0)


class LatencyServerService(ILatencyServerService, IReceiveListener, AbstractService):
    def __init__(
            self,
            arguments: LatencyServerArguments,
            clock: IClock,
            routed_receive_service: RoutedReceiveService,
            routed_send_service: RoutedSendService,
            service_manager: AbstractServiceManager,
    ) -> None:
        super().__init__(service_manager)
        self.arguments = arguments
        self.clock = clock
        self.interruptable_sleep = InterruptableSleep(clock)
        self.listeners: Listeners[ILatencyListener] = Listeners()
        self.received_messages: IQueue[Optional[LatencyMessage]] = IdealQueue()
        self.send_service = routed_send_service.create_send_service(message_types.latency)
        self.thread = Thread(target=self.run)
        routed_receive_service.create_receive_service(message_types.latency).add_receive_listener(self)

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.received_messages.put(LatencyMessage.decode(message))

    def run(self) -> None:
        sent_message = LatencyMessage(0)
        while True:
            try:
                sent_message.increment()
                sent_timestamp = self.clock.get_seconds()
                self.send_service.send(sent_message.encode())
                while True:
                    timeout_seconds = sent_timestamp + self.arguments.latency_check_interval_seconds.get(
                    ) - self.clock.get_seconds()
                    if timeout_seconds <= 0:
                        raise Empty()
                    received_message = self.clock.get_from_queue(self.received_messages, timeout_seconds)
                    received_timestamp = self.clock.get_seconds()
                    if received_message is None:
                        return
                    if received_message.number == sent_message.number:
                        break
                self.listeners.for_each(lambda listener: listener.on_latency_available(
                    received_timestamp - sent_timestamp))
                if not self.interruptable_sleep.sleep(self.arguments.latency_check_interval_seconds.get()):
                    return
            except Empty:
                self.listeners.for_each(lambda listener: listener.on_latency_timeout())

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.interruptable_sleep.interrupt()
        self.received_messages.put(None)

    def add_latency_listener(self, latency_listener: ILatencyListener) -> LatencyServerService:
        self.listeners.add_listener(latency_listener)
        return self



from ..argument import IArguments
from ..gpio.IGpioServerService import IGpioServerService
from ..latency.ILatencyListener import ILatencyListener
from ..latency.ILatencyServerService import ILatencyServerService
from ..timeout.ITimeoutReceiveListener import ITimeoutReceiveListener
from ..timeout.TimeoutReceiveService import TimeoutReceiveService
from ..util.Atomic import Atomic
from ..util.Lazy import Lazy
from ..util.util import min_max


class Throttle:
    def __init__(self, max_latency_seconds: float, max_speed: int) -> None:
        self.max_latency_seconds = max_latency_seconds
        self.max_speed = max_speed

    def get_throttled_speed(self, speed: int) -> int:
        return min_max(speed, -self.max_speed, self.max_speed)


class ThrottleServerArguments(IArguments):
    def __init__(self) -> None:
        self.throttle_upgrade_factor = Lazy(lambda store: 0.9)
        self.throttles = Lazy(lambda store: [Throttle(0.25, 100), Throttle(0.5, 50), Throttle(1.0, 10)])
        self.worst_throttle = Lazy(lambda store: Throttle(0.1, 0))


class ThrottleServerService(
    ILatencyListener,
    ITimeoutReceiveListener,
):
    def __init__(
            self,
            arguments: ThrottleServerArguments,
            gpio_service: IGpioServerService,
            latency_service: ILatencyServerService,
            timeout_service: TimeoutReceiveService,
    ) -> None:
        self.arguments = arguments
        self.current_throttle = Atomic(arguments.worst_throttle.get())
        self.gpio_service = gpio_service
        latency_service.add_latency_listener(self)
        self.timeout_service = timeout_service.add_timeout_listener(self)

    def on_latency_available(self, latency: float) -> None:
        new_throttle = self.arguments.worst_throttle.get()
        for throttle in self.arguments.throttles.get():
            if latency <= throttle.max_latency_seconds:
                new_throttle = throttle
                break
        with self.current_throttle as (current_throttle, set_current_throttle):
            if new_throttle.max_latency_seconds < current_throttle.max_latency_seconds:
                if latency < new_throttle.max_latency_seconds * self.arguments.throttle_upgrade_factor.get():
                    set_current_throttle(new_throttle)
            else:
                set_current_throttle(new_throttle)
        self.timeout_service.set_timeout(new_throttle.max_latency_seconds)

    def on_latency_timeout(self) -> None:
        with self.current_throttle as (_, set_current_throttle):
            set_current_throttle(self.arguments.worst_throttle.get())

    def on_timeout_receive(self) -> None:
        with self.current_throttle as (_, set_current_throttle):
            set_current_throttle(self.arguments.worst_throttle.get())
        self.gpio_service.reset()

    def update(self, motor_value: int, steering_value: int) -> None:
        self.gpio_service.update(self.current_throttle.get().get_throttled_speed(motor_value), steering_value)



from __future__ import annotations
from typing import Dict
from .IRoutedReceiveListener import IRoutedReceiveListener
from .RoutedMessage import RoutedMessage
from ..receive import IReceiveListener, IReceiveService
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Listeners import Listeners


class ReceiveListener(IRoutedReceiveListener):
    def __init__(self, listener: IReceiveListener) -> None:
        self.listener = listener

    def on_routed_receive(self, message_type: int, message: bytes, details: ConnectionDetails) -> None:
        self.listener.on_receive(message, details)


class ReceiveService(IReceiveService):
    def __init__(self, message_type: int, routed_receive_service: RoutedReceiveService) -> None:
        self.message_type = message_type
        self.routed_receive_service = routed_receive_service

    def add_receive_listener(self, listener: IReceiveListener) -> ReceiveService:
        self.routed_receive_service.add_receive_listener(self.message_type, ReceiveListener(listener))
        return self


class RoutedReceiveService(IReceiveListener):
    def __init__(self, receive_service: IReceiveService) -> None:
        self.received_listeners: Dict[int, Listeners[IRoutedReceiveListener]] = dict()
        receive_service.add_receive_listener(self)

    def add_receive_listener(self, message_type: int, listener: IRoutedReceiveListener) -> RoutedReceiveService:
        if message_type > 255:
            raise RuntimeError(f'Illegal message type: {message_type} > 255.')
        if message_type not in self.received_listeners:
            self.received_listeners[message_type] = Listeners()
        self.received_listeners[message_type].add_listener(listener)
        return self

    def create_receive_service(self, message_type: int) -> IReceiveService:
        return ReceiveService(message_type, self)

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        routed_message = RoutedMessage.decode(message)
        self.received_listeners[routed_message.message_type].for_each(
            lambda listener: listener.on_routed_receive(routed_message.message_type, routed_message.payload, details))



from __future__ import annotations
from abc import ABC, abstractmethod
from ..util.ConnectionDetails import ConnectionDetails


class IRoutedReceiveListener(ABC):
    @abstractmethod
    def on_routed_receive(self, message_type: int, message: bytes, details: ConnectionDetails) -> None:
        """"""



from __future__ import annotations
from abc import ABC, abstractmethod
from .util.ConnectionDetails import ConnectionDetails


class IReceiveListener(ABC):
    @abstractmethod
    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        """"""


class IReceiveService(ABC):
    @abstractmethod
    def add_receive_listener(self, listener: IReceiveListener) -> IReceiveService:
        """"""



from logging import getLogger
from typing import Callable, Generic, List, TypeVar

T = TypeVar('T')


class Listeners(Generic[T]):
    def __init__(self) -> None:
        self.listeners: List[T] = []
        self.logger = getLogger(__name__)

    def add_listener(self, listener: T) -> None:
        self.listeners.append(listener)

    def for_each(self, callback: Callable[[T], None]) -> None:
        for listener in self.listeners:
            try:
                callback(listener)
            except BaseException as exception:
                self.logger.exception(f'Exception thrown in listener callback:', exc_info=exception)



from ..argument import AnyArguments, IArguments, ValueArgument
from ..constants import default_discovery_port
from ..util.Lazy import Lazy


class DiscoveryCommonArguments(IArguments):
    def __init__(self) -> None:
        self.port = Lazy(lambda store: default_discovery_port)

    def get_arguments(self) -> AnyArguments:
        return [ValueArgument(self.port, '--discovery-port', int, 'Discovery broadcast port.')]



from abc import ABC, abstractmethod
from typing import Union


class _MessageTypes:
    def __init__(self) -> None:
        message_type_generator = iter(range(2**8))
        self.car = next(message_type_generator)
        self.custom = next(message_type_generator)
        self.latency = next(message_type_generator)
        self.stop = next(message_type_generator)
        self.video = next(message_type_generator)


message_types = _MessageTypes()


class IMessage(ABC):
    @abstractmethod
    def encode(self) -> bytes:
        """"""


Payload = Union[bytes, IMessage]


def encode_payload(payload: Payload) -> bytes:
    return payload.encode() if isinstance(payload, IMessage) else payload



from __future__ import annotations
from abc import ABC, abstractmethod
from socket import SHUT_RDWR
from types import TracebackType
from typing import Callable, Optional, Tuple, Type, Union


class ISocket(ABC):
    @abstractmethod
    def __enter__(self) -> ISocket:
        """"""

    @abstractmethod
    def __exit__(
            self, exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        """"""

    @abstractmethod
    def accept(self) -> Tuple[ISocket, Tuple[str, int]]:
        """"""

    @abstractmethod
    def bind(self, address: Tuple[str, int]) -> None:
        """"""

    @abstractmethod
    def close(self) -> None:
        """"""

    @abstractmethod
    def connect(self, address: Tuple[str, int]) -> None:
        """"""

    @abstractmethod
    def getsockname(self) -> Tuple[str, int]:
        """"""

    @abstractmethod
    def listen(self, backlog: int) -> None:
        """"""

    @abstractmethod
    def recv(self, bufsize: int) -> bytes:
        """"""

    @abstractmethod
    def recvfrom(self, bufsize: int) -> Tuple[bytes, Optional[Tuple[str, int]]]:
        """"""

    @abstractmethod
    def sendall(self, data: bytes) -> None:
        """"""

    @abstractmethod
    def sendto(self, data: bytes, address: Tuple[str, int]) -> int:
        """"""

    @abstractmethod
    def setsockopt(self, level: int, optname: int, value: Union[int, bytes]) -> None:
        """"""

    @abstractmethod
    def settimeout(self, value: float) -> None:
        """"""

    @abstractmethod
    def shutdown(self, how: int) -> None:
        """"""

    def shutdown_guaranteed(self, on_os_error: Callable[[OSError], None] = lambda os_error: None) -> None:
        try:
            # Notifies recvfrom() that the socket is shutdown, but shutdown on UDP sockets always throws an OSError,
            # but recvfrom() returns (b'', None) in such case.
            self.shutdown(SHUT_RDWR)
        except OSError as os_error:
            on_os_error(os_error)


class ISocketFactory(ABC):
    @abstractmethod
    def socket(self, family: int, type_: int, proto: int = -1) -> ISocket:
        """"""



from .ExpireMessage import ExpireMessage, highest_message_number
from ..send import ISendService
from ..util.Atomic import Atomic


class ExpireSendService(ISendService):
    def __init__(self, send_service: ISendService) -> None:
        self.send_service = send_service
        self.last_message_number = Atomic(1)

    def send(self, message: bytes) -> None:
        with self.last_message_number as (last_message_number, set_last_message_number):
            message_number = last_message_number
            set_last_message_number((last_message_number + 1) % highest_message_number)
        self.send_service.send(ExpireMessage(message_number, message).encode())



from __future__ import annotations
from struct import pack, unpack
from typing import cast, Tuple
from ..message import IMessage

format_string = '<bb'
format_type = Tuple[int, int]


class CarMessage(IMessage):
    def __init__(self, speed: int, steering: int) -> None:
        self._speed = speed
        self._steering = steering

    def update(self, speed: int, steering: int) -> CarMessage:
        self._speed = speed
        self._steering = steering
        return self

    @property
    def speed(self) -> int:
        return self._speed

    @property
    def steering(self) -> int:
        return self._steering

    def encode(self) -> bytes:
        return pack(format_string, self._speed, self._steering)

    @staticmethod
    def decode(message: bytes) -> CarMessage:
        speed, steering = cast(format_type, unpack(format_string, message))
        return CarMessage(speed, steering)



from __future__ import annotations
from logging import getLogger
from typing import Tuple
from .CarMessage import CarMessage
from ..argument import AnyArguments, IArguments, ValueArgument
from ..clock import IClock
from ..expire.ExpireSendService import ExpireSendService
from ..message import message_types
from ..reliable import IReliableConnectListener, IReliableDisconnectListener, IReliableService
from ..routed.RoutedSendService import RoutedSendService
from ..priority.PrioritySendService import PrioritySendService
from ..service import AbstractServiceManager
from ..timeout.TimeoutSendService import TimeoutSendService
from ..util.Lazy import Lazy
from ..util.ConnectionDetails import ConnectionDetails


class CarClientArguments(IArguments):
    def __init__(self) -> None:
        self.priority = Lazy(lambda store: 128)
        self.send_timeout_seconds = Lazy(lambda store: 0.125)

    def get_arguments(self) -> AnyArguments:
        return [ValueArgument(self.priority, '--priority', int, 'Priority of client.')]


class CarClientService(IReliableConnectListener, IReliableDisconnectListener):
    def __init__(
            self,
            arguments: CarClientArguments,
            clock: IClock,
            reliable_service: IReliableService,
            routed_send_service: RoutedSendService,
            service_manager: AbstractServiceManager,
    ) -> None:
        self._car_message = CarMessage(0, 0)
        self.logger = getLogger(__name__)
        self.timeout_send_service = TimeoutSendService(
            clock,
            arguments.send_timeout_seconds.get(),
            ExpireSendService(
                PrioritySendService(
                    arguments.priority.get(),
                    routed_send_service.create_send_service(message_types.car))),
            service_manager,
        )
        reliable_service.add_reliable_connect_listener(self).add_reliable_disconnect_listener(self)

    def on_reliable_connect(self, details: ConnectionDetails) -> None:
        self.timeout_send_service.set_and_send_immediately(lambda: self._car_message.encode())

    def on_reliable_disconnect(self, details: ConnectionDetails) -> None:
        self.timeout_send_service.set_and_send_immediately(lambda: None)
        with self.timeout_send_service.send_count as (send_count, _):
            self.logger.info(f'Client statistics: sent={send_count}')

    def update(self, speed: int, steering: int) -> Tuple[int, int]:
        self.timeout_send_service.set_and_send_immediately(lambda: self._car_message.update(speed, steering).encode())
        return speed, steering



from typing import Any, Tuple, Type, TypeVar
from ..constants import buffer_size

V1 = TypeVar('V1')
V2 = TypeVar('V2')


class BufferSizeExceededError(RuntimeError):
    def __init__(self, length: int) -> None:
        super().__init__(f'Buffer size exceeded: {length} > {buffer_size}')


def check_message_length(length: int) -> None:
    if length > buffer_size:
        raise BufferSizeExceededError(length)


class UnexpectedTypeError(TypeError):
    def __init__(self, subject: Any, expected: str) -> None:
        super().__init__(f'Expected type {expected}, got {subject}')


def check_type(subject: Any, type_: Type[V1]) -> V1:
    if not isinstance(subject, type_):
        raise UnexpectedTypeError(subject, str(type_))
    return subject


def check_type_tuple2(subject: Any, type1: Type[V1], type2: Type[V2]) -> Tuple[V1, V2]:
    def create_error() -> UnexpectedTypeError:
        return UnexpectedTypeError(subject, f'({type1}, {type2})')

    if not isinstance(subject, tuple):
        raise create_error()
    if not len(subject) == 2:
        raise create_error()
    value1, value2 = subject
    if not isinstance(value1, type1):
        raise create_error()
    if not isinstance(value2, type2):
        raise create_error()
    return value1, value2



from __future__ import annotations
from logging import getLogger
from math import inf
from socket import AF_INET, SOCK_DGRAM
from threading import Thread
from typing import Optional, Tuple
from ..clock import IClock
from ..constants import buffer_size
from ..log.util import log_method_call
from ..receive import IReceiveListener
from ..service import AbstractService, AbstractServiceManager
from ..socket_ import ISocket, ISocketFactory
from ..unreliable import IUnreliableOsErrorListener, IUnreliableReceiveListener, IUnreliableReceiveSendService
from ..util import checking
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Listeners import Listeners
from ..util.Placeholder import Placeholder


class UdpService(AbstractService, IUnreliableReceiveSendService):
    def __init__(
            self,
            clock: IClock,
            is_server: bool,
            server_address: Placeholder[Tuple[str, int]],
            service_manager: AbstractServiceManager,
            socket_factory: ISocketFactory,
            unreliable_timeout_seconds: float = inf,
    ) -> None:
        super().__init__(service_manager)
        logger = getLogger(__name__)
        self.clock = clock
        self.is_server = is_server
        self.os_error_listeners: Listeners[IUnreliableOsErrorListener] = Listeners()
        self.receiver_address: Optional[Tuple[str, int]] = None
        self.last_received_seconds = -inf
        self.received_listeners: Listeners[IReceiveListener] = Listeners()
        self.server_address = server_address
        self.service_manager = service_manager
        self.should_run = True
        self.socket: Placeholder[ISocket] = Placeholder()
        self.socket_factory = socket_factory
        self.thread = Thread(target=log_method_call(logger, self.run))
        self.unreliable_received_listeners: Listeners[IUnreliableReceiveListener] = Listeners()
        self.unreliable_timeout_seconds = unreliable_timeout_seconds

    def add_unreliable_os_error_listener(self, listener: IUnreliableOsErrorListener) -> UdpService:
        self.os_error_listeners.add_listener(listener)
        return self

    def add_receive_listener(self, listener: IReceiveListener) -> UdpService:
        self.received_listeners.add_listener(listener)
        return self

    def add_unreliable_receive_listener(
            self,
            listener: IUnreliableReceiveListener
    ) -> UdpService:
        self.unreliable_received_listeners.add_listener(listener)
        return self

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def run(self) -> None:
        with self.socket.set(self.socket_factory.socket(AF_INET, SOCK_DGRAM)) as socket:
            server_address = self.server_address.get_eventually(self.service_manager).get_blocking()
            if server_address is None:
                return
            if self.is_server:
                socket.bind(server_address)
            else:
                self.receiver_address = server_address

            while self.should_run:
                try:
                    (message, receiver_address) = socket.recvfrom(buffer_size)
                    if receiver_address is None:
                        break
                    self.last_received_seconds = self.clock.get_seconds()
                    connection_details = ConnectionDetails(socket.getsockname(), receiver_address)
                    self.receiver_address = receiver_address
                    self.received_listeners.for_each(
                        lambda listener: listener.on_receive(message, connection_details))
                    self.unreliable_received_listeners.for_each(
                        lambda listener: listener.on_unreliable_receive(message, connection_details))
                except OSError as os_error:
                    self.os_error_listeners.for_each(
                        lambda listener: listener.on_unreliable_os_error(os_error))

    def send(self, message: bytes) -> None:
        checking.check_message_length(len(message))
        receiver_address = self.receiver_address
        if receiver_address is None:
            return
        if self.last_received_seconds + self.unreliable_timeout_seconds < self.clock.get_seconds():
            return
        self.socket.get().sendto(message, receiver_address)

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        socket = self.socket.get_optional_and_clear()
        if socket is not None:
            socket.shutdown_guaranteed()



from __future__ import annotations
from threading import Event
from typing import Generic, Optional, TypeVar
from .IPromise import IPromise
from ..service import AbstractStartedService, AbstractServiceManager

T = TypeVar('T')


class PromiseService(Generic[T], IPromise[T], AbstractStartedService):
    def __init__(self, service_manager: AbstractServiceManager) -> None:
        super().__init__(service_manager)
        self._event = Event()
        self._value: Optional[T] = None

    def get_blocking(self) -> Optional[T]:
        self._event.wait()
        return self._value

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        return self._event.is_set()

    def reject(self) -> IPromise[T]:
        self._event.set()
        return self

    def resolve(self, value: T) -> IPromise[T]:
        self._value = value
        self._event.set()
        return self

    def stop_service(self) -> None:
        self.reject()



from __future__ import annotations
from logging import getLogger
from .CarMessage import CarMessage
from .NetworkStatisticsMessage import NetworkStatisticsMessage
from ..argument import IArguments
from ..clock import IClock
from ..expire.ExpireReceiveService import ExpireReceiveService
from ..expire.IExpireReceiveListener import IExpireReceiveListener
from ..gpio.IGpioServerService import IGpioServerService
from ..latency.ILatencyServerService import ILatencyServerService
from ..message import message_types
from ..priority.PriorityReceiveService import PriorityReceiveService
from ..receive import IReceiveListener
from ..reliable import IReliableConnectListener, IReliableDisconnectListener, IReliableService
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..service import AbstractServiceManager
from ..throttle.ThrottleServerService import ThrottleServerArguments, ThrottleServerService
from ..timeout.ITimeoutReceiveListener import ITimeoutReceiveListener
from ..timeout.TimeoutReceiveService import TimeoutReceiveService
from ..util.Atomic import Atomic
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Lazy import Lazy


class CarServerArguments(IArguments):
    def __init__(self) -> None:
        self.priority_timeout_seconds = Lazy(lambda store: 1.0)
        self.throttle = ThrottleServerArguments()


class CarServerService(
    IExpireReceiveListener,
    IReceiveListener,
    IReliableConnectListener,
    IReliableDisconnectListener,
    ITimeoutReceiveListener,
):
    def __init__(
            self,
            arguments: CarServerArguments,
            clock: IClock,
            gpio_service: IGpioServerService,
            latency_service: ILatencyServerService,
            reliable_service: IReliableService,
            routed_receive_service: RoutedReceiveService,
            service_manager: AbstractServiceManager,
    ) -> None:
        self.logger = getLogger(__name__)
        self.expire_service = ExpireReceiveService(PriorityReceiveService(
            clock, routed_receive_service.create_receive_service(message_types.car),
            service_manager, arguments.priority_timeout_seconds.get())).add_expire_listener(self)
        self.network_statistics = Atomic(NetworkStatisticsMessage(0, 0, 0))
        self.throttle_service = ThrottleServerService(
            arguments.throttle, gpio_service, latency_service, TimeoutReceiveService(
                clock, self.expire_service, service_manager).add_receive_listener(self).add_timeout_listener(self))
        reliable_service.add_reliable_connect_listener(self).add_reliable_disconnect_listener(self)

    def on_expire_receive(self, message: bytes, details: ConnectionDetails) -> None:
        with self.network_statistics as (network_statistics, _):
            network_statistics.expire_receive_count += 1

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        with self.network_statistics as (network_statistics, _):
            network_statistics.receive_count += 1
        car_message = CarMessage.decode(message)
        self.throttle_service.update(car_message.speed, car_message.steering)

    def on_reliable_connect(self, details: ConnectionDetails) -> None:
        with self.expire_service.last_message_number as (_, set_last_message_number):
            set_last_message_number(0)
        with self.network_statistics as (network_statistics, _):
            network_statistics.expire_receive_count = 0
            network_statistics.receive_count = 0
            network_statistics.timeout_receive_count = 0

    def on_reliable_disconnect(self, details: ConnectionDetails) -> None:
        with self.network_statistics as (network_statistics, _):
            self.logger.info(
                f'Client statistics: receive={network_statistics.receive_count}, '
                f'expire={network_statistics.expire_receive_count}, timeout={network_statistics.timeout_receive_count}')

    def on_timeout_receive(self) -> None:
        with self.network_statistics as (network_statistics, _):
            network_statistics.timeout_receive_count += 1



from ..argument import AnyArguments, IArguments, ValueArgument
from ..constants import default_server_port, discover_server_ip
from ..util.Lazy import Lazy


class ClientArguments(IArguments):
    def __init__(self) -> None:
        self.reconnect_seconds = Lazy(lambda store: 1.0)
        self.server_ip = Lazy(lambda store: discover_server_ip)
        self.server_port = Lazy(lambda store: default_server_port)
        self.server_address = Lazy(lambda store: (self.server_ip(store), self.server_port(store)))

    def get_arguments(self) -> AnyArguments:
        return [
            ValueArgument(self.server_ip, '--server-ip', str, 'TCP/UDP server IP.'),
            ValueArgument(self.server_port, '--server-port', int, 'TCP/UDP server port.'),
        ]



from queue import Empty, Queue
from sys import version_info
from typing import Generic, Optional, TypeVar
from ..queue_ import IQueue

T = TypeVar('T')


class IdealQueue(Generic[T], IQueue[T]):
    def __init__(self) -> None:
        self._call_task_done = False
        if version_info >= (3, 8):
            queue: Queue[T] = Queue()
        else:
            queue = Queue()
        self._queue = queue

    def empty(self) -> bool:
        return self._queue.empty()

    def get(self, block: bool = True, timeout: Optional[float] = None) -> T:
        call_task_done = self._call_task_done
        self._call_task_done = True
        if call_task_done:
            self._queue.task_done()
        try:
            return self._queue.get(block=block, timeout=timeout)
        except Empty as exception:
            self._call_task_done = False
            raise exception

    def join(self) -> None:
        self._queue.join()

    def put(self, item: T) -> None:
        self._queue.put(item)



from .Atomic import Atomic


class AtomicInteger(Atomic[int]):
    def __init__(self, value: int) -> None:
        super().__init__(value)

    def increment(self, amount: int = 1) -> None:
        with self as (value, set_value):
            set_value(value + amount)



from __future__ import annotations
from queue import Empty
from threading import Thread
from typing import Optional, Union
from .ITimeoutReceiveListener import ITimeoutReceiveListener
from .ITimeoutReceiveService import ITimeoutReceiveService
from ..clock import IClock
from ..queue_ import IQueue
from ..receive import IReceiveListener, IReceiveService
from ..service import AbstractService, AbstractServiceManager
from ..util.ConnectionDetails import ConnectionDetails
from ..util.IdealQueue import IdealQueue
from ..util.Listeners import Listeners


class ReceiveEvent:
    pass


class SetTimeoutEvent:
    def __init__(self, timeout_seconds: Optional[float]) -> None:
        self.timeout_seconds = timeout_seconds


class StopEvent:
    pass


class TimeoutReceiveService(AbstractService, IReceiveListener, IReceiveService, ITimeoutReceiveService):
    def __init__(
            self,
            clock: IClock,
            receive_service: IReceiveService,
            service_manager: AbstractServiceManager,
            timeout_seconds: Optional[float] = None,
    ) -> None:
        super().__init__(service_manager)
        self.clock = clock
        self.event_bus: IQueue[Union[ReceiveEvent, SetTimeoutEvent, StopEvent]] = IdealQueue()
        self.receive_listeners: Listeners[IReceiveListener] = Listeners()
        self.thread = Thread(target=self.run)
        self.timeout_listeners: Listeners[ITimeoutReceiveListener] = Listeners()
        self.timeout_seconds = timeout_seconds
        receive_service.add_receive_listener(self)

    def add_receive_listener(self, listener: IReceiveListener) -> TimeoutReceiveService:
        self.receive_listeners.add_listener(listener)
        return self

    def add_timeout_listener(self, timeout_listener: ITimeoutReceiveListener) -> TimeoutReceiveService:
        self.timeout_listeners.add_listener(timeout_listener)
        return self

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.event_bus.put(ReceiveEvent())
        self.receive_listeners.for_each(lambda listener: listener.on_receive(message, details))

    def run(self) -> None:
        while True:
            try:
                waiting_since_seconds = self.clock.get_seconds()
                while True:
                    if self.timeout_seconds is None:
                        event = self.event_bus.get()
                    else:
                        remaining_seconds = waiting_since_seconds + self.timeout_seconds - self.clock.get_seconds()
                        if remaining_seconds <= 0:
                            raise Empty()
                        event = self.clock.get_from_queue(self.event_bus, remaining_seconds)
                    if isinstance(event, ReceiveEvent):
                        break
                    elif isinstance(event, SetTimeoutEvent):
                        self.timeout_seconds = event.timeout_seconds
                    elif isinstance(event, StopEvent):
                        return
            except Empty:
                self.timeout_listeners.for_each(lambda listener: listener.on_timeout_receive())

    def set_timeout(self, timeout_seconds: float) -> None:
        self.event_bus.put(SetTimeoutEvent(timeout_seconds))

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.event_bus.put(StopEvent())



from abc import ABC, abstractmethod
from ..util.ConnectionDetails import ConnectionDetails


class IExpireReceiveListener(ABC):
    @abstractmethod
    def on_expire_receive(self, message: bytes, details: ConnectionDetails) -> None:
        """"""



from threading import Lock
from types import TracebackType
from typing import Callable, Generic, Optional, Tuple, Type, TypeVar

T = TypeVar('T')


class Atomic(Generic[T]):
    def __init__(self, value: T) -> None:
        self._lock = Lock()
        self._value = value

    def __enter__(self) -> Tuple[T, Callable[[T], T]]:
        self._lock.__enter__()
        return self._value, self._set_value

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        self._lock.__exit__(exc_type, exc_val, exc_tb)

    def get(self) -> T:
        return self._value

    def _set_value(self, value: T) -> T:
        self._value = value
        return value



class VideoSettings:
    def __init__(self, bit_rate: int, fps: int, height: int, width: int) -> None:
        self.bit_rate = bit_rate
        self.fps = fps
        self.height = height
        self.width = width



from ..argument import AnyArguments, IArguments, ValueArgument
from ..constants import default_server_port
from ..util.Lazy import Lazy


class ServerArguments(IArguments):
    def __init__(self) -> None:
        self.listen_ip = Lazy(lambda store: '0.0.0.0')
        self.listen_port = Lazy(lambda store: default_server_port)
        self.listen_address = Lazy(lambda store: (self.listen_ip(store), self.listen_port(store)))
        self.unreliable_timeout_seconds = Lazy(lambda store: 60.0)

    def get_arguments(self) -> AnyArguments:
        return [
            ValueArgument(self.listen_ip, '--listen-ip', str, 'TCP/UDP listen IP.'),
            ValueArgument(self.listen_port, '--listen-port', int, 'TCP/UDP listen port.'),
            ValueArgument(
                self.unreliable_timeout_seconds, '--unreliable-timeout-seconds', float,
                'Stop sending unreliable messages to last known client after receiving nothing for this amount of '
                'seconds.'
            ),
        ]



from subprocess import Popen
from typing import IO, Optional, Sequence
from ..process import FILE, IProcess


class Process(IProcess):
    def __init__(self, args: Sequence[str], stdin: Optional[FILE], stdout: Optional[FILE]) -> None:
        self.process = Popen(args, stdin=stdin, stdout=stdout)

    def get_stdout(self) -> IO[bytes]:
        if self.process.stdout is None:
            raise RuntimeError('Unexpected None. Is stdout set to PIPE?')
        return self.process.stdout

    def terminate(self) -> None:
        self.process.terminate()

    def wait(self, timeout: Optional[float]) -> None:
        self.process.wait(timeout)



from datetime import datetime
from threading import Event, Lock, Thread
from typing import Optional, TypeVar
from ..clock import IClock
from ..queue_ import IQueue

T = TypeVar('T')


class _RTC(IClock):
    """
    Real-time clock
    """
    def acquire_lock(self, lock: Lock, timeout_seconds: Optional[float] = None) -> bool:
        if timeout_seconds is None:
            return lock.acquire()
        else:
            return lock.acquire(timeout=timeout_seconds)

    def get_from_queue(self, queue: IQueue[T], timeout_seconds: Optional[float] = None) -> T:
        return queue.get(timeout=timeout_seconds)

    def get_seconds(self) -> float:
        return datetime.utcnow().timestamp()

    def join_thread(self, thread: Thread, timeout_seconds: Optional[float] = None) -> None:
        thread.join(timeout_seconds)

    def notify(self) -> None:
        pass

    def wait_for_event(self, event: Event, timeout_seconds: Optional[float] = None) -> bool:
        return event.wait(timeout_seconds)


rtc = _RTC()



from __future__ import annotations
from logging import getLogger
from socket import AF_INET, SOCK_STREAM
from threading import Thread
from typing import Optional, Tuple
from ..clock import IClock
from ..constants import buffer_size
from ..log.util import log_method_call
from ..receive import IReceiveListener
from ..reliable import IReliableReceiveListener, IReliableReceiveSendService, IReliableConnectListener
from ..reliable import IReliableDisconnectListener, IReliableOsErrorListener
from ..service import AbstractService, AbstractServiceManager
from ..socket_ import ISocket, ISocketFactory
from ..util import checking
from ..util.ConnectionDetails import ConnectionDetails
from ..util.InterruptableSleep import InterruptableSleep
from ..util.Listeners import Listeners
from ..util.Placeholder import Placeholder


class TcpClientService(AbstractService, IReliableReceiveSendService):
    def __init__(
            self,
            clock: IClock,
            reconnect_seconds: float,
            server_address: Placeholder[Tuple[str, int]],
            service_manager: AbstractServiceManager,
            socket_factory: ISocketFactory,
    ) -> None:
        super().__init__(service_manager)
        self.connected_listeners: Listeners[IReliableConnectListener] = Listeners()
        self.disconnected_listeners: Listeners[IReliableDisconnectListener] = Listeners()
        self.interruptable_sleep = InterruptableSleep(clock)
        self.logger = getLogger(__name__)
        self.os_error_listeners: Listeners[IReliableOsErrorListener] = Listeners()
        self.received_listeners: Listeners[IReceiveListener] = Listeners()
        self.reconnect_seconds = reconnect_seconds
        self.reliable_received_listeners: Listeners[IReliableReceiveListener] = Listeners()
        self.server_address = server_address
        self.service_manager = service_manager
        self.should_run = True
        self.socket: Placeholder[ISocket] = Placeholder()
        self.socket_factory = socket_factory
        self.thread = Thread(target=log_method_call(self.logger, self.run))

    def add_reliable_connect_listener(self, listener: IReliableConnectListener) -> TcpClientService:
        self.connected_listeners.add_listener(listener)
        return self

    def add_reliable_disconnect_listener(self, listener: IReliableDisconnectListener) -> TcpClientService:
        self.disconnected_listeners.add_listener(listener)
        return self

    def add_reliable_os_error_listener(self, listener: IReliableOsErrorListener) -> TcpClientService:
        self.os_error_listeners.add_listener(listener)
        return self

    def add_receive_listener(self, listener: IReceiveListener) -> TcpClientService:
        self.received_listeners.add_listener(listener)
        return self

    def add_reliable_receive_listener(
            self,
            listener: IReliableReceiveListener
    ) -> TcpClientService:
        self.reliable_received_listeners.add_listener(listener)
        return self

    def get_service_name(self) -> str:
        return __name__

    def get_own_address(self) -> Tuple[str, int]:
        return self.socket.get().getsockname()

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def run_with_server_address(self, server_address: Tuple[str, int]) -> None:
        while self.should_run:
            try:
                with self.socket.set(self.socket_factory.socket(AF_INET, SOCK_STREAM)) as socket:
                    socket.connect(server_address)
                    connection_details = ConnectionDetails(socket.getsockname(), server_address)
                    self.connected_listeners.for_each(
                        lambda listener: listener.on_reliable_connect(connection_details))
                    try:
                        while self.should_run:
                            message = socket.recv(buffer_size)
                            if len(message) == 0:  # ... means that peer has gracefully shutdown connection
                                break
                            self.received_listeners.for_each(
                                lambda listener: listener.on_receive(message, connection_details))
                            self.reliable_received_listeners.for_each(
                                lambda listener: listener.on_reliable_receive(message, connection_details))
                    finally:
                        self.disconnected_listeners.for_each(
                            lambda listener: listener.on_reliable_disconnect(connection_details))
            except OSError as os_error:
                # Is thrown if connection could not be established or got lost
                self.os_error_listeners.for_each(lambda listener: listener.on_reliable_os_error(os_error))
            if self.should_run:
                self.socket.clear()
                self.interruptable_sleep.sleep(self.reconnect_seconds)

    def run(self) -> None:
        server_address = self.server_address.get_eventually(self.service_manager).get_blocking()
        if server_address is not None:
            self.run_with_server_address(server_address)

    def send(self, message: bytes) -> None:
        checking.check_message_length(len(message))
        self.socket.get().sendall(message)

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        self.interruptable_sleep.interrupt()
        socket = self.socket.get_optional_and_clear()
        if socket is not None:
            socket.shutdown_guaranteed(
                lambda os_error: self.logger.info(
                    'TCP socket was already shutdown or connection was never established.'))



from __future__ import annotations
from ..constants import discovery_magic_request_bytes


class DiscoveryRequestMessage:
    @staticmethod
    def encode() -> bytes:
        return discovery_magic_request_bytes

    @staticmethod
    def is_valid(message: bytes) -> bool:
        return message == discovery_magic_request_bytes



from __future__ import annotations
from typing import Callable, Generic, Optional, TypeVar

T = TypeVar('T')


class Lazy(Generic[T]):
    def __init__(self, compute_default: Callable[[bool], T]) -> None:
        self._compute_default = compute_default
        self._is_computing = False
        self._value: Optional[T] = None

    def __call__(self, store: bool) -> T:
        if self._value is None:
            if self._is_computing:
                raise RuntimeError('Circular dependency detected. See stacktrace.')
            self._is_computing = True
            value = self._compute_default(store)
            self._is_computing = False
            if store:
                self._value = value
            return value
        else:
            return self._value

    def get_default(self) -> T:
        return self(False)

    def get(self) -> T:
        return self(True)

    def set(self, value: T) -> None:
        self._value = value



from typing import Generic, List, Optional, TypeVar
from .Atomic import Atomic

T = TypeVar('T')


class FirstInLastOut(Generic[T]):
    """
    Thread-safe first-in-last-out
    """
    def __init__(self) -> None:
        self._filo: Atomic[List[T]] = Atomic([])

    def append(self, item: T) -> None:
        with self._filo as (filo, _):
            filo.append(item)

    def pop(self) -> Optional[T]:
        with self._filo as (filo, _):
            if len(filo) == 0:
                return None
            return filo.pop()



from socket import socket
from typing import Tuple
from .ClientArguments import ClientArguments
from ..car.CarClientService import CarClientArguments, CarClientService
from ..clock import IClock
from ..constants import discover_server_ip
from ..discovery.DiscoveryClientService import DiscoveryClientArguments, DiscoveryClientService
from ..gstreamer.GStreamerClientService import GStreamerClientArguments, GStreamerClientService
from ..latency.LatencyClientService import LatencyClientService
from ..log.LogArguments import LogArguments
from ..reliable import IReliableReceiveSendService
from ..routed.RoutedSendService import RoutedSendService
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..service import AbstractServiceManager
from ..socket_ import ISocketFactory
from ..tcp.TcpClientService import TcpClientService
from ..udp.UdpService import UdpService
from ..unreliable import IUnreliableReceiveSendService
from ..util.Lazy import Lazy
from ..util.MultiServiceManager import MultiServiceManager
from ..util.Placeholder import Placeholder
from ..util.RTC import rtc
from ..util.Socket import SocketFactory


class Client:
    def __init__(self) -> None:
        self.car_arguments = CarClientArguments()
        self.client_arguments = ClientArguments()
        self.discovery_arguments = DiscoveryClientArguments()
        self.gstreamer_arguments = GStreamerClientArguments()
        self.log_arguments = LogArguments()
        self.clock: Lazy[IClock] = Lazy(lambda store: rtc)
        self.socket_factory: Lazy[ISocketFactory] = Lazy(lambda store: SocketFactory(socket))
        self.service_manager: Lazy[AbstractServiceManager] = Lazy(lambda store: MultiServiceManager(self.clock(store)))
        self.discovery_service = Lazy(lambda store: DiscoveryClientService(
            self.discovery_arguments, self.service_manager(store), self.socket_factory(store)))

        def create_server_address(store: bool) -> Placeholder[Tuple[str, int]]:
            if self.client_arguments.server_address(store)[0] == discover_server_ip:
                return self.discovery_service(store).server_address
            else:
                return Placeholder(self.client_arguments.server_address(store))

        self.server_address = Lazy(create_server_address)
        self.reliable_service: Lazy[IReliableReceiveSendService] = Lazy(lambda store: TcpClientService(
            self.clock(store), self.client_arguments.reconnect_seconds(store), self.server_address(store),
            self.service_manager(store), self.socket_factory(store)))
        self.unreliable_service: Lazy[IUnreliableReceiveSendService] = Lazy(lambda store: UdpService(
            self.clock(store), False, self.server_address(store), self.service_manager(store),
            self.socket_factory(store)))
        self.reliable_routed_receive_service = Lazy(lambda store: RoutedReceiveService(
            self.reliable_service(store)))
        self.reliable_routed_send_service = Lazy(lambda store: RoutedSendService(self.reliable_service(store)))
        self.unreliable_routed_receive_service = Lazy(lambda store: RoutedReceiveService(
            self.unreliable_service(store)))
        self.unreliable_routed_send_service = Lazy(lambda store: RoutedSendService(self.unreliable_service(store)))
        self.car_service = Lazy(lambda store: CarClientService(
            self.car_arguments, self.clock(store), self.reliable_service(store),
            self.unreliable_routed_send_service(store), self.service_manager(store)))
        self.gstreamer_service = Lazy(lambda store: GStreamerClientService(
            self.gstreamer_arguments, self.reliable_routed_receive_service(store),
            self.reliable_routed_send_service(store)))
        self.latency_service = Lazy(lambda store: LatencyClientService(
            self.unreliable_routed_receive_service(store), self.unreliable_routed_send_service(store)))
